import json
import os
import re
import shutil
from pathlib import Path

import pytest
from eth_utils import to_hex
from ethpm_types import Compiler, ContractType, PackageManifest, Source

from ape import Project
from ape.contracts import ContractContainer
from ape.exceptions import ConfigError, ProjectError
from ape.logging import LogLevel
from ape.managers.project import MultiProject
from ape.utils import create_tempdir
from ape_pm.project import BrownieProject, FoundryProject


@pytest.fixture
def contract_type():
    return make_contract("FooContractFromManifest")


@pytest.fixture
def manifest(contract_type):
    return make_manifest(contract_type)


@pytest.fixture
def contract_block_hash(eth_tester_provider, vyper_contract_instance):
    block_number = vyper_contract_instance.creation_metadata.block
    return to_hex(eth_tester_provider.get_block(block_number).hash)


@pytest.fixture
def project_from_manifest(manifest):
    return Project.from_manifest(manifest)


@pytest.fixture
def tmp_project(project):
    with project.isolate_in_tempdir() as tmp_project:
        yield tmp_project


@pytest.fixture(scope="module")
def foundry_toml():
    return """
[profile.default]
src = 'src'
out = 'out'
libs = ['lib']
solc = "0.8.18"
evm_version = 'cancun'
via_ir = true

remappings = [
    'forge-std/=lib/forge-std/src/',
    '@openzeppelin/=lib/openzeppelin-contracts/',
]
""".lstrip()


def make_contract(name: str = "test") -> ContractType:
    return ContractType.model_validate(
        {
            "contractName": name,
            "sourceId": f"contracts/{name}.json",
            "abi": [],
        }
    )


def make_manifest(*contracts: ContractType, include_contract_type: bool = True) -> PackageManifest:
    sources = {
        ct.source_id: Source(content=ct.model_dump_json(by_alias=True, mode="json"))
        for ct in contracts
    }
    model: dict = {"sources": sources}
    if include_contract_type:
        contract_types = {c.name: c for c in contracts}
        model["contractTypes"] = contract_types

    return PackageManifest.model_validate(model)


def test_path(smaller_project):
    assert smaller_project.path is not None


def test_path_configured():
    """
    Simulating package structures like snekmate.
    """
    madeup_name = "snakemate"
    with create_tempdir(name=madeup_name) as temp_dir:
        subdir = temp_dir / "src"
        contracts_folder = subdir / madeup_name
        contracts_folder.mkdir(parents=True)
        contract = contracts_folder / "snake.json"
        abi = [{"name": "foo", "type": "fallback", "stateMutability": "nonpayable"}]
        contract.write_text(json.dumps(abi), encoding="utf8")

        snakemate = Project(
            temp_dir, config_override={"base_path": "src", "contracts_folder": madeup_name}
        )
        assert snakemate.name == madeup_name
        assert snakemate.path == subdir
        assert snakemate.contracts_folder == contracts_folder

        # The repr should show `/snakemate` and not `/snakemate/src/`.
        assert re.match(r"<ProjectManager [\w|/]*/snakemate>", repr(snakemate))

        actual = snakemate.load_contracts()
        assert "snake" in actual
        assert actual["snake"].source_id == f"{madeup_name}/snake.json"


def test_name(smaller_project):
    assert smaller_project.name == smaller_project.path.name


def test_name_from_config(project):
    with project.temp_config(name="foo-bar"):
        assert project.name == "foo-bar"


def test_repr():
    with create_tempdir() as tmpdir:
        project = Project(tmpdir)
        actual = repr(project)

    expected = f"<ProjectManager {tmpdir}>"
    assert actual == expected


@pytest.mark.parametrize("name", ("contracts", "contracts"))
def test_contracts_folder_from_config(smaller_project, name):
    with smaller_project.temp_config(contracts_folder=name):
        assert smaller_project.contracts_folder == smaller_project.path / name


def test_contracts_folder_same_as_root_path(smaller_project):
    with smaller_project.temp_config(contracts_folder="."):
        assert smaller_project.contracts_folder == smaller_project.path


def test_contracts_folder_deduced(tmp_project):
    new_project_path = tmp_project.path / "new"
    new_project_path.mkdir()
    contracts_folder = new_project_path / "contracts"
    contracts_folder.mkdir()
    contract = contracts_folder / "tryme.json"
    abi = [{"name": "foo", "type": "fallback", "stateMutability": "nonpayable"}]
    contract.write_text(json.dumps(abi), encoding="utf8")
    new_project = Project(new_project_path)
    actual = new_project.contracts_folder
    assert actual == contracts_folder


def test_reconfigure(small_temp_project):
    small_temp_project.reconfigure(compile={"exclude": ["first", "second"]})
    assert {"first", "second"}.issubset(set(small_temp_project.config.compile.exclude))


def test_isolate_in_tempdir(smaller_project):
    with smaller_project.isolate_in_tempdir() as tmp_project:
        assert tmp_project.path != smaller_project.path
        assert tmp_project.in_tempdir


def test_isolate_in_tempdir_does_not_alter_sources(smaller_project):
    # First, create a bad source.
    with smaller_project.temp_config(contracts_folder="build"):
        new_src = smaller_project.contracts_folder / "newsource.json"
        new_src.parent.mkdir(exist_ok=True, parents=True)
        new_src.write_text("this is not json, oops")
        smaller_project.sources.refresh()  # Only need to be called when run with other tests.

        try:
            with smaller_project.isolate_in_tempdir() as tmp_project:
                # The new (bad) source should be in the temp project.
                actual = {**(tmp_project.manifest.sources or {})}
        finally:
            new_src.unlink()
            smaller_project.sources.refresh()

        # Ensure "newsource" did not persist in the in-memory manifest.
        assert "build/newsource.json" in actual
        assert "build/newsource.json" not in (smaller_project.manifest.sources or {})


def test_in_tempdir(smaller_project):
    assert not Project(Path(__file__).parent).in_tempdir
    with smaller_project.isolate_in_tempdir() as tmp_project:
        assert tmp_project.in_tempdir


def test_getattr(smaller_project):
    actual = smaller_project.Other
    assert type(actual) is ContractContainer


def test_getattr_not_exists(smaller_project):
    expected = (
        r"'LocalProject' object has no attribute 'nope'\. Also checked extra\(s\) 'contracts'\."
    )
    with pytest.raises(AttributeError, match=expected) as err:
        _ = smaller_project.nope

    # Was the case where the last entry was from Ape's basemodel stuff.
    # Now, it points at the project manager last.
    assert "ape/managers/project.py:" in repr(err.traceback[-1])


def test_getattr_detects_changes(small_temp_project):
    # For this test, ensure we use a JSON/ABI contract so it is easier
    # to change its contents.
    source_id = small_temp_project.Other.contract_type.source_id
    new_abi = {
        "inputs": [],
        "name": "retrieve",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    }
    content = json.dumps([new_abi])
    path = small_temp_project.sources.lookup(source_id)
    assert path

    path.unlink(missing_ok=True)
    path.write_text(content, encoding="utf8")
    # Should have re-compiled.
    contract = small_temp_project.Other
    assert "retrieve" in contract.contract_type.methods


def test_getattr_empty_contract(small_temp_project):
    """
    Tests against a condition where would infinitely compile.
    """
    source_id = small_temp_project.Other.contract_type.source_id
    path = small_temp_project.sources.lookup(source_id)
    path.unlink(missing_ok=True)
    path.write_text("", encoding="utf8")
    # Should have re-compiled.
    contract = small_temp_project.Other
    assert not contract.contract_type.methods


def test_getattr_same_name_as_source_file(small_temp_project):
    new_file = small_temp_project.contracts_folder / "NewContract.sol"
    content = """
    // SPDX-License-Identifier: LGPL-3.0-only
    pragma solidity >=0.7.0 <0.8.0;
    contract DifferentName {}
    """
    new_file.write_text(content, encoding="utf8")
    expected = (
        r"'LocalProject' object has no attribute 'NewContract'\. "
        r"Also checked extra\(s\) 'contracts'\. "
        r"However, there is a source file named 'NewContract\.sol'\. "
        r"This file may not be compiling \(see error above\), "
        r"or maybe you meant to reference a contract name from this source file\?*"
    )
    with pytest.raises(AttributeError, match=expected):
        _ = small_temp_project.NewContract


@pytest.mark.parametrize("iypthon_attr_name", ("_repr_mimebundle_", "_ipython_display_"))
def test_getattr_ipython(smaller_project, iypthon_attr_name):
    # Remove contract types, if there is any.
    smaller_project.manifest.contract_types = {}
    getattr(smaller_project, iypthon_attr_name)
    # Prove it did not compile looking for these names.
    assert not smaller_project.manifest.contract_types


def test_getattr_ipython_canary_check(tmp_project):
    tmp_project.manifest.contract_types = {}

    with pytest.raises(AttributeError):
        _ = tmp_project._ipython_canary_method_should_not_exist_

    assert not tmp_project.manifest.contract_types  # Didn't compile


def test_getitem(smaller_project):
    actual = smaller_project["Other"]
    assert type(actual) is ContractContainer


def test_meta(smaller_project):
    meta_config = {
        "meta": {
            "authors": ["Apealicious Jones"],
            "license": "MIT",
            "description": "Zoologist meme protocol",
            "keywords": ["Indiana", "Knight's Templar"],
            "links": {"apeworx.io": "https://apeworx.io"},
        }
    }
    with smaller_project.temp_config(**meta_config):
        assert smaller_project.meta.authors == ["Apealicious Jones"]
        assert smaller_project.meta.license == "MIT"
        assert smaller_project.meta.description == "Zoologist meme protocol"
        assert smaller_project.meta.keywords == ["Indiana", "Knight's Templar"]
        assert len(smaller_project.meta.links) == 1
        assert f"{smaller_project.meta.links['apeworx.io']}" == "https://apeworx.io/"


def test_extract_manifest(smaller_project, vyper_contract_instance):
    contract_type = vyper_contract_instance.contract_type
    smaller_project.manifest.contract_types = {contract_type.name: contract_type}
    smaller_project.deployments.track(vyper_contract_instance, allow_dev=True)

    manifest = smaller_project.extract_manifest()
    assert type(manifest) is PackageManifest
    assert manifest.meta == smaller_project.meta
    assert "manifest-dependency" in (manifest.dependencies or {})
    bip122_chain_id = to_hex(smaller_project.provider.get_block(0).hash)
    expected_uri = f"blockchain://{bip122_chain_id[2:]}"
    for key in manifest.deployments or {}:
        if key.startswith(expected_uri):
            return

    assert False, "Failed to find expected deployment URI"


def test_extract_manifest_when_sources_missing(empty_project):
    """
    Show that if a source is missing, it is OK. This happens when changing branches
    after compiling and contracts are only present on one of the branches.
    """
    contract = make_contract("notreallyhere")
    empty_project.manifest.contract_types = {contract.name: contract}
    manifest = empty_project.extract_manifest()

    # Source is skipped because missing.
    assert "notreallyhere" not in manifest.contract_types


def test_extract_manifest_excludes_cache(empty_project):
    cachefile = empty_project.contracts_folder / ".cache" / "CacheFile.json"
    cachefile2 = empty_project.contracts_folder / ".cache" / "subdir" / "Cache2.json"
    cachefile2.parent.mkdir(parents=True)
    cachefile.write_text("Doesn't matter", encoding="utf8")
    cachefile2.write_text("Doesn't matter", encoding="utf8")
    manifest = empty_project.extract_manifest()
    assert isinstance(manifest, PackageManifest)
    assert ".cache/CacheFile.json" not in (manifest.sources or {})
    assert ".cache/subdir/CacheFile.json" not in (manifest.sources or {})


def test_extract_manifest_compiles(smaller_project):
    smaller_project.manifest.contract_types = {}  # Not compiled.
    actual = smaller_project.extract_manifest()
    assert actual.contract_types  # Fails if empty


def test_extract_manifest_from_manifest_project(project_from_manifest):
    project_from_manifest.manifest.contract_types = {}  # Not compiled.
    manifest = project_from_manifest.extract_manifest()
    assert "FooContractFromManifest" in manifest.contract_types


def test_exclusions(smaller_project):
    exclusions = ["Other.json", "*Excl*"]
    exclude_config = {"compile": {"exclude": exclusions}}
    with smaller_project.temp_config(**exclude_config):
        for exclusion in exclusions:
            assert exclusion in smaller_project.exclusions


def test_update_manifest(empty_project):
    compiler = Compiler(name="comp", version="1.0.0", contractTypes=["foo.txt"])
    empty_project.update_manifest(compilers=[compiler])
    actual = empty_project.manifest.compilers
    assert actual == [compiler]

    empty_project.update_manifest(name="test", version="1.0.0")
    assert empty_project.manifest.name == "test"
    assert empty_project.manifest.version == "1.0.0"

    # The compilers should not have changed.
    actual = empty_project.manifest.compilers
    assert actual == [compiler]


def test_load_contracts(small_temp_project):
    contracts = small_temp_project.load_contracts()
    assert small_temp_project.manifest_path.is_file()
    assert len(contracts) > 0
    contracts_forced = small_temp_project.load_contracts(use_cache=False)
    assert len(contracts_forced) > 0

    # Delete a file and ensure it does not show up in dict.
    try:
        contract_to_rm = small_temp_project.contracts["Other"]
    except KeyError:
        existing_contracts = ",".join([k for k in small_temp_project.contracts.keys()])
        pytest.fail(f"Contract named 'Other' not found. Existing contracts: {existing_contracts}")
        return

    contract_path = small_temp_project.sources.lookup(contract_to_rm.source_id)
    contract_path.unlink()
    contracts = small_temp_project.load_contracts()
    assert contract_to_rm.name not in contracts


def test_load_contracts_detect_change(small_temp_project, ape_caplog):
    path = small_temp_project.contracts_folder / "Other.json"
    content = path.read_text(encoding="utf8")
    assert "foo" in content, "Unexpected content"

    # Must be compiled first.
    with ape_caplog.at_level(LogLevel.INFO):
        contracts = small_temp_project.load_contracts(use_cache=False)
        assert "Other" in contracts
        ape_caplog.assert_last_log("Compiling")

        ape_caplog.clear()

        # No logs as it doesn't need to re-compile.
        small_temp_project.load_contracts()
        assert not ape_caplog.head

        # Make a change to the file.
        new_content = content.replace("foo", "bar")
        assert "bar" in new_content, "Test setup failed. Unexpected file content."
        path.unlink()
        path.write_text(new_content, encoding="utf8")

        # Prove re-compiles.
        contracts = small_temp_project.load_contracts()
        assert "Other" in contracts
        ape_caplog.assert_last_log("Compiling")


def test_load_contracts_after_deleting_same_named_contract(empty_project, compilers, mock_compiler):
    """
    Tests against a scenario where you:

    1. Add and compile a contract
    2. Delete that contract
    3. Add a new contract with same name somewhere else

    Test such that we are able to compile successfully and not get a misleading
    collision error from deleted files.
    """
    init_contract = empty_project.contracts_folder / "foo.__mock__"
    init_contract.parent.mkdir(parents=True, exist_ok=True)
    init_contract.write_text("LALA", encoding="utf8")
    compilers.registered_compilers[".__mock__"] = mock_compiler
    result = empty_project.load_contracts()
    assert "foo" in result

    # Goodbye.
    init_contract.unlink()

    # Since we are changing files mid-session, we need to refresh the project.
    # Typically, users don't have to do this.
    empty_project.refresh_sources()

    result = empty_project.load_contracts()
    assert "foo" not in result  # Was deleted.
    # Also ensure it is gone from paths.
    assert "foo.__mock__" not in [x.name for x in empty_project.sources.paths]

    # Create a new contract with the same name.
    new_contract = empty_project.contracts_folder / "bar.__mock__"
    new_contract.write_text("BAZ", encoding="utf8")
    empty_project.refresh_sources()

    mock_compiler.overrides = {"contractName": "foo"}
    result = empty_project.load_contracts()
    assert "foo" in result


def test_load_contracts_output_abi(smaller_project):
    cfg = {"output_extra": ["ABI"]}
    with smaller_project.temp_config(compile=cfg):
        _ = smaller_project.load_contracts(use_cache=False)
        abi_folder = smaller_project.manifest_path.parent / "abi"
        assert abi_folder.is_dir()
        files = [x for x in abi_folder.iterdir()]
        assert len(files) > 0
        for file in files:
            assert file.suffix == ".json"

        # Ensure is usable.
        data = json.loads(file.read_text(encoding="utf8"))
        assert isinstance(data, list)
        assert len(data) >= 1
        # There was bug where this was a str.
        assert isinstance(data[0], dict)


def test_load_contracts_use_cache(mocker, smaller_project):
    """
    Showing the 'use_cache=bool' kwarg works.
    """
    compile_spy = mocker.spy(smaller_project.contracts, "_compile")

    smaller_project.manifest.contract_types = {}  # Force initial compile.
    contracts = smaller_project.load_contracts(use_cache=True)
    assert "Other" in contracts  # Other.json contract.
    assert compile_spy.call_args_list[-1][-1]["use_cache"] is True

    # Show they get added to the manifest.
    assert "Other" in smaller_project.manifest.contract_types

    # Show we can use the cache again (no compiling!)
    contracts = smaller_project.load_contracts(use_cache=True)
    assert "Other" in contracts  # Other.json contract.
    assert compile_spy.call_args_list[-1][-1]["use_cache"] is True

    # Show force-recompiles.
    contracts = smaller_project.load_contracts(use_cache=False)
    assert "Other" in contracts  # Other.json contract.
    assert compile_spy.call_args_list[-1][-1]["use_cache"] is False


def test_manifest_path(smaller_project):
    assert smaller_project.manifest_path == smaller_project.path / ".build" / "__local__.json"


def test_clean(small_temp_project):
    small_temp_project.load_contracts()
    assert small_temp_project.manifest_path.is_file()

    small_temp_project.clean()
    assert not small_temp_project.manifest_path.is_file()
    assert small_temp_project._manifest.contract_types is None
    assert small_temp_project.sources._path_cache is None
    assert small_temp_project._manifest.compilers is None


def test_unpack(small_temp_project):
    # Create a none-contract file.
    non_contract_name = "ThisIsNotAContract.txt"
    path = small_temp_project.contracts_folder / non_contract_name
    path.write_text("not a contract", encoding="utf8")

    with create_tempdir() as path:
        small_temp_project.unpack(path)
        assert (path / "contracts" / "Other.json").is_file()

        # Show that even non-contracts end up in the unpacked destination.
        assert (path / "contracts" / non_contract_name).is_file()


def test_unpack_includes_build_file(small_temp_project):
    buildfile = small_temp_project.path / ".build" / "__local__.json"
    buildfile.parent.mkdir(parents=True, exist_ok=True)
    buildfile.write_text("{}", encoding="utf8")

    with create_tempdir() as path:
        small_temp_project.unpack(path)
        assert (path / ".build" / "__local__.json").is_file()


def test_unpack_includes_interfaces():
    iname = "unp_interface"
    with create_tempdir() as path:
        interfaces = path / "interfaces"
        interfaces.mkdir(parents=True, exist_ok=True)
        interface = interfaces / f"{iname}.json"
        interface.write_text("{}", encoding="utf8")

        project = Project(path)
        with create_tempdir() as new_path:
            project.unpack(new_path)
            expected_interface = new_path / "interfaces" / f"{iname}.json"
            assert expected_interface.is_file()


def test_unpack_includes_interfaces_when_part_of_contracts():
    iname = "unp_interface"
    with create_tempdir() as path:
        interfaces = path / "contracts" / "interfaces"
        interfaces.mkdir(parents=True, exist_ok=True)
        interface = interfaces / f"{iname}.json"
        interface.write_text("{}", encoding="utf8")

        project = Project(path)
        with create_tempdir() as new_path:
            project.unpack(new_path)
            expected_interface = new_path / "contracts" / "interfaces" / f"{iname}.json"
            assert expected_interface.is_file()


def test_add_compiler_data(project_with_dependency_config):
    project_with_dependency_config.clean()

    # NOTE: Using different project than default to lessen
    #   chance of race-conditions from multiprocess test runners.
    project = project_with_dependency_config

    # Load contracts so that any compilers that may exist are present.
    project.load_contracts()

    compiler = Compiler(
        name="comp",
        version="1.0.0",
        contractTypes=["foo"],
        settings={"outputSelection": {"path/to/Foo.sol": "*"}},
    )

    # Ensure types are in manifest for type-source-id lookup.
    bar = ContractType(contractName="bar", sourceId="path/to/Bar.vy")
    foo = ContractType(contractName="foo", sourceId="path/to/Foo.sol")
    project._manifest = PackageManifest(
        contractTypes={"bar": bar, "foo": foo},
        sources={"path/to/Bar.vy": Source(), "path/to/Foo.vy": Source()},
    )
    project._contracts = project._manifest.contract_types
    assert project._manifest.contract_types, "Setup failed - need manifest contract types"

    # Add twice to show it's only added once.
    project.add_compiler_data([compiler])
    project.add_compiler_data([compiler])

    if compiler.name not in [x.name for x in project.manifest.compilers]:
        names = [x.name for x in project.manifest.compilers]
        pytest.fail(f"Missing expected compiler with name '{compiler.name}'. Names: {names}")
    if compiler.version not in [x.version for x in project.manifest.compilers]:
        pytest.fail(f"Missing expected compiler with version '{compiler.version}'")

    # NOTE: `add_compiler_data()` will not override existing compilers.
    #   Use `update_cache()` for that.
    compiler_2 = Compiler(
        name="test",
        version="2.0.0",
        contractTypes=["bar", "stay"],
        settings={"outputSelection": {"path/to/Bar.vy": "*", "stay.vy": "*"}},
    )
    project.add_compiler_data([compiler_2])
    assert project.manifest.compilers == [compiler, compiler_2]

    # NOTE: Has same contract as compiler 2 and thus replaces the contract.
    compiler_3 = Compiler(
        name="test",
        version="3.0.0",
        contractTypes=["bar"],
        settings={"outputSelection": {"path/to/Bar.vy": "*"}},
    )
    project.add_compiler_data([compiler_3])
    comp = [c for c in project.manifest.compilers if c.name == "test" and c.version == "2.0.0"][0]
    assert "bar" not in comp.contractTypes
    assert "path/to/Bar.vy" not in comp.settings["outputSelection"]
    new_comp = [c for c in project.manifest.compilers if c.name == "test" and c.version == "3.0.0"][
        0
    ]
    assert "bar" in new_comp.contractTypes
    assert "path/to/Bar.vy" in new_comp.settings["outputSelection"]

    # Show that compilers without contract types go away.
    (compiler_3.contractTypes or []).append("stay")
    project.add_compiler_data([compiler_3])
    comp_check = [
        c for c in project.manifest.compilers if c.name == "test" and c.version == "2.0.0"
    ]
    assert not comp_check

    # Show error on multiple of same compiler.
    compiler_4 = Compiler(name="test123", version="3.0.0", contractTypes=["bar"])
    compiler_5 = compiler_4.model_copy(update={"contractTypes": ["baz"]})
    with pytest.raises(ProjectError, match=r".*was given multiple of the same compiler.*"):
        project.add_compiler_data([compiler_4, compiler_5])

    # Show error when contract type collision (only happens with inputs, else latter replaces).
    compiler_5.contractTypes = ["bar"]
    compiler_5.name = f"{compiler_5.name}new"
    with pytest.raises(ProjectError, match=r".*'bar' collision across compilers.*"):
        project.add_compiler_data([compiler_4, compiler_5])


def test_project_api_foundry_and_ape_config_found(foundry_toml):
    """
    If a project is both a Foundry project and an Ape project,
    ensure both configs are honored.
    """
    with create_tempdir() as tmpdir:
        foundry_cfg_file = tmpdir / "foundry.toml"
        foundry_cfg_file.write_text(foundry_toml, encoding="utf8")

        ape_cfg_file = tmpdir / "ape-config.yaml"
        ape_cfg_file.write_text("name: testfootestfootestfoo", encoding="utf8")

        temp_project = Project(tmpdir)
        assert isinstance(temp_project.project_api, MultiProject)

        # This came from the ape config file.
        assert temp_project.config.name == "testfootestfootestfoo"

        # This came from the foundry toml file.
        assert temp_project.config.contracts_folder == "src"
        assert len(temp_project.config.solidity.import_remapping) > 0


def test_get_contract(smaller_project):
    actual = smaller_project.get_contract("Other")
    assert isinstance(actual, ContractContainer), f"{type(actual)}"
    assert actual.contract_type.name == "Other"

    # Ensure manifest is only loaded once by none-ing out the path.
    # Otherwise, this can be a MAJOR performance hit.
    manifest_path = smaller_project.manifest_path
    smaller_project.manifest_path = None
    try:
        actual = smaller_project.get_contract("Other")
        assert isinstance(actual, ContractContainer)
        assert actual.contract_type.name == "Other"
    finally:
        smaller_project.manifest_path = manifest_path


def test_get_contract_not_exists(smaller_project):
    actual = smaller_project.get_contract("this is not a contract")
    assert actual is None


class TestProject:
    """
    All tests related to ``ape.Project``.
    """

    def test_init(self, with_dependencies_project_path):
        # Purpose not using `project_with_contracts` fixture.
        project = Project(with_dependencies_project_path)

        # NOTE: Using tempdir to avoid clashing with other tests during x-dist.
        with project.isolate_in_tempdir() as temp_project:
            assert project.path == with_dependencies_project_path
            project.manifest_path.unlink(missing_ok=True)

            #  Re-init to show it doesn't create the manifest file.
            _ = Project(temp_project.path)

    def test_init_invalid_config(self):
        here = os.curdir
        with create_tempdir() as temp_dir:
            cfgfile = temp_dir / "ape-config.yaml"
            # Name is invalid!
            cfgfile.write_text("name:\n  {asdf}")

            os.chdir(temp_dir)
            expected = r"[.\n]*Input should be a valid string\n-->1: name:\n   2:   {asdf}[.\n]*"
            weird_project = Project(temp_dir)
            try:
                with pytest.raises(ConfigError, match=expected):
                    _ = weird_project.path
            finally:
                os.chdir(here)

    def test_config_override(self, with_dependencies_project_path):
        contracts_folder = with_dependencies_project_path / "my_contracts"
        config = {"contracts_folder": contracts_folder.name}
        project = Project(with_dependencies_project_path, config_override=config)
        assert project.contracts_folder == contracts_folder

    def test_from_manifest(self, manifest):
        # Purposely not using `project_from_manifest` fixture.
        project = Project.from_manifest(manifest)
        assert isinstance(project, Project)
        assert project.manifest == manifest

    def test_from_manifest_contracts_iter(self, contract_type, project_from_manifest):
        actual = set(iter(project_from_manifest.contracts))
        assert actual == {"FooContractFromManifest"}

    def test_from_manifest_getattr(self, contract_type, project_from_manifest):
        expected = ContractContainer(contract_type)
        actual = project_from_manifest.FooContractFromManifest
        assert isinstance(actual, ContractContainer)
        assert actual == expected

    def test_from_manifest_getitem(self, contract_type, project_from_manifest):
        expected = ContractContainer(contract_type)
        assert project_from_manifest["FooContractFromManifest"] == expected

    def test_from_manifest_load_contracts(self, contract_type):
        """
        Show if contract-types are missing but contracts set,
        compiling will add contract-types.
        """
        manifest = make_manifest(contract_type, include_contract_type=False)
        project = Project.from_manifest(manifest)
        assert not project.manifest.contract_types, "Setup failed"

        # Returns containers, not types.
        actual = project.load_contracts()
        assert actual[contract_type.name].contract_type == contract_type

        # Also, show it got set on the manifest.
        assert project.manifest.contract_types == {contract_type.name: contract_type}

    def test_from_python_library(self):
        # web3py as an ape-project dependency.
        web3 = Project.from_python_library("web3")
        assert "site-packages" in str(web3.path)
        assert web3.path.name == "web3"

    def test_hash(self, with_dependencies_project_path, project_from_manifest):
        """
        Show we can use projects in sets.
        """
        project_0 = Project(with_dependencies_project_path)
        project_1 = Project.from_python_library("web3")
        project_2 = project_from_manifest

        # Show we can use in sets.
        project_set = {project_0, project_1, project_2}
        assert len(project_set) == 3

        # Show we can use as dict-keys:
        project_dict = {project_0: 123}
        assert project_dict[project_0] == 123


class TestBrownieProject:
    """
    Tests related to the brownie implementation of the ProjectAPI.
    """

    @pytest.fixture
    def brownie_project(self, base_projects_directory):
        project_path = base_projects_directory / "BrownieProject"
        return BrownieProject(path=project_path)

    def test_extract_config(self, config, brownie_project):
        config = brownie_project.extract_config()

        # Ensure contracts_folder works.
        assert config.contracts_folder == "contractsrenamed"

        # Ensure Solidity and dependencies configuration mapped correctly
        assert config.solidity.version == "0.6.12"

        # NOTE: `contracts/` is not part of the import key as it is
        # usually included in the import statements.
        assert [str(x) for x in config.solidity.import_remapping] == [
            "@openzeppelin=openzeppelin/3.1.0"
        ]
        assert config.dependencies[0]["name"] == "openzeppelin"
        assert config.dependencies[0]["github"] == "OpenZeppelin/openzeppelin-contracts"
        assert config.dependencies[0]["version"] == "3.1.0"


class TestFoundryProject:
    @pytest.fixture
    def mock_github(self, mocker):
        return mocker.MagicMock()

    @pytest.fixture(scope="class")
    def gitmodules(self):
        return """
[submodule "lib/forge-std"]
    path = lib/forge-std
    url = https://github.com/foundry-rs/forge-std
    branch = v1.5.2
[submodule "lib/openzeppelin-contracts"]
    path = lib/openzeppelin-contracts
    url = https://github.com/OpenZeppelin/openzeppelin-contracts
    release = v4.9.5
    branch = v4.9.5
[submodule "lib/erc4626-tests"]
    path = lib/erc4626-tests
    url = https://github.com/a16z/erc4626-tests.git
""".lstrip().replace("    ", "\t")

    def test_extract_config(self, foundry_toml, gitmodules, mock_github):
        with create_tempdir() as temp_dir:
            cfg_file = temp_dir / "foundry.toml"
            cfg_file.write_text(foundry_toml, encoding="utf8")
            gitmodules_file = temp_dir / ".gitmodules"
            gitmodules_file.write_text(gitmodules, encoding="utf8")
            temp_project = Project(temp_dir)

            api = temp_project.project_api
            mock_github.get_repo.return_value = {"default_branch": "main"}
            api._github_client = mock_github  # type: ignore
            assert isinstance(api, FoundryProject)

            # Ensure solidity config migrated.
            actual = temp_project.config.model_dump(
                by_alias=True
            )  # Is result of ``api.extract_config()``.
            assert actual["contracts_folder"] == "src"
            assert "solidity" in actual, "Solidity failed to migrate"
            actual_sol = actual["solidity"]
            assert actual_sol["import_remapping"] == [
                "@openzeppelin=src/.cache/openzeppelin/v4.9.5/",
                "forge-std=src/.cache/forge-std/v1.5.2/src",
            ]
            assert actual_sol["version"] == "0.8.18"
            assert actual_sol["evm_version"] == "cancun"
            assert actual_sol["via_ir"] is True

            # Ensure dependencies migrated from .gitmodules.
            assert "dependencies" in actual, "Dependencies failed to migrate"
            actual_dependencies = actual["dependencies"]
            expected_dependencies = [
                {"github": "foundry-rs/forge-std", "name": "forge-std", "ref": "v1.5.2"},
                {
                    "github": "OpenZeppelin/openzeppelin-contracts",
                    "name": "openzeppelin",
                    "version": "v4.9.5",
                },
                {"github": "a16z/erc4626-tests", "name": "erc4626-tests", "ref": "main"},
            ]
            assert actual_dependencies == expected_dependencies


class TestSourceManager:
    def test_lookup(self, smaller_project):
        source_id = smaller_project.Other.contract_type.source_id
        path = smaller_project.sources.lookup(source_id)
        assert path == smaller_project.path / source_id

    def test_lookup_same_source_id_as_local_project(self, smaller_project):
        """
        Tests against a bug where if the source ID of the project matched
        a file in the local project, it would mistakenly return the path
        to the local file instead of the project's file.
        """
        source_id = smaller_project.contracts["Other"].source_id
        path = smaller_project.sources.lookup(source_id)
        assert path.is_file(), "Test path does not exist."

        cfg = {"contracts_folder": smaller_project.config.contracts_folder}
        with Project.create_temporary_project(config_override=cfg) as tmp_project:
            new_source = tmp_project.path / source_id
            new_source.parent.mkdir(parents=True, exist_ok=True)
            new_source.write_text(path.read_text(encoding="utf8"), encoding="utf8")

            actual = tmp_project.sources.lookup(source_id)
            assert actual is not None
            expected = tmp_project.path / source_id
            assert actual == expected

    def test_lookup_missing_extension(self, smaller_project):
        source_id = smaller_project.Other.contract_type.source_id
        source_id_wo_ext = ".".join(source_id.split(".")[:-1])
        path = smaller_project.sources.lookup(source_id_wo_ext)
        assert path == smaller_project.path / source_id

    def test_lookup_mismatched_extension(self, smaller_project):
        source_id = smaller_project.Other.contract_type.source_id
        source_id = source_id.replace(Path(source_id).suffix, ".js")
        path = smaller_project.sources.lookup(source_id)
        assert path is None

    def test_lookup_closest_match(self, smaller_project):
        source_path = smaller_project.contracts_folder / "Other.json"
        temp_dir_a = smaller_project.contracts_folder / "temp"
        temp_dir_b = temp_dir_a / "tempb"
        nested_source_a = temp_dir_a / "Other.json"
        nested_source_b = temp_dir_b / "Other.json"

        def clean():
            # NOTE: Will also delete temp_dir_b.
            if temp_dir_a.is_dir():
                shutil.rmtree(temp_dir_a)

        clean()

        # NOTE: Will also make temp_dir_a.
        temp_dir_b.mkdir(parents=True)

        try:
            # Duplicate contract so that there are multiple with the same name.
            for nested_src in (nested_source_a, nested_source_b):
                nested_src.touch()
                nested_src.write_text(source_path.read_text(), encoding="utf8")

            # Top-level match.
            for base in (source_path, str(source_path), "Other", "Other.json"):
                # Using stem in case it returns `Contract.__mock__`, which is
                # added / removed as part of other tests (running x-dist).
                assert smaller_project.sources.lookup(base).stem == source_path.stem, (
                    f"Failed to lookup {base}"
                )

            # Nested: 1st level
            for closest in (
                nested_source_a,
                str(nested_source_a),
                "temp/Other",
                "temp/Other.json",
            ):
                actual = smaller_project.sources.lookup(closest)
                assert actual

                expected = nested_source_a
                # Using stem in case it returns `Contract.__mock__`, which is
                # added / removed as part of other tests (running x-dist).
                assert actual.stem == expected.stem, f"Failed to lookup {closest}"

            # Nested: 2nd level
            for closest in (
                nested_source_b,
                str(nested_source_b),
                "temp/tempb/Other",
                "temp/tempb/Other.json",
            ):
                actual = smaller_project.sources.lookup(closest)
                expected = nested_source_b

                # Using stem in case it returns `Contract.__mock__`, which is
                # added / removed as part of other tests (running x-dist).
                assert actual.stem == expected.stem, f"Failed to lookup {closest}"

        finally:
            clean()

    def test_lookup_not_found(self, smaller_project):
        assert smaller_project.sources.lookup("madeup.json") is None

    def test_lookup_missing_contracts_prefix(self, smaller_project):
        """
        Show we can exclude the `contracts/` prefix in a source ID.
        """
        actual_from_str = smaller_project.sources.lookup("Other.json")
        actual_from_path = smaller_project.sources.lookup(Path("Other.json"))
        expected = smaller_project.contracts_folder / "Other.json"
        assert actual_from_str == actual_from_path == expected
        assert actual_from_str.is_absolute()
        assert actual_from_path.is_absolute()

    def test_paths_exclude(self, smaller_project):
        exclude_config = {"compile": {"exclude": ["Other.json"]}}
        with smaller_project.temp_config(**exclude_config):
            # Show default excludes also work, such as a .DS_Store file.
            ds_store = smaller_project.contracts_folder / ".DS_Store"
            ds_store.write_bytes(b"asdfasf")

            # Show anything in compiler-cache is ignored.
            cache = smaller_project.contracts_folder / ".cache"
            cache.mkdir(exist_ok=True)
            random_file = cache / "dontmindme.json"
            random_file.write_text("what, this isn't json?!", encoding="utf8")

            path_ids = {
                f"{smaller_project.contracts_folder.name}/{src.name}"
                for src in smaller_project.sources.paths
            }
            excluded = {".DS_Store", "Other.json", ".cache/dontmindme.json"}
            for actual in (path_ids, smaller_project.sources):
                for exclusion in excluded:
                    expected = f"{smaller_project.contracts_folder.name}/{exclusion}"
                    assert expected not in actual

    def test_is_excluded(self, smaller_project):
        exclude_cfg = {"compile": {"exclude": ["exclude_dir/*", "Excl*.json"]}}
        source_ids = ("contracts/exclude_dir/UnwantedContract.json", "contracts/Exclude.json")
        with smaller_project.temp_config(**exclude_cfg):
            for source_id in source_ids:
                path = smaller_project.path / source_id
                assert smaller_project.sources.is_excluded(path)

    def test_items(self, smaller_project):
        actual = list(smaller_project.sources.items())
        assert len(actual) > 0
        assert isinstance(actual[0], tuple)
        assert "contracts/Other.json" in [x[0] for x in actual]
        assert isinstance(actual[0][1], Source)

    def test_keys(self, smaller_project):
        actual = list(smaller_project.sources.keys())
        assert "contracts/Other.json" in actual

    def test_values(self, smaller_project):
        actual = list(smaller_project.sources.values())
        assert all(isinstance(x, Source) for x in actual)


class TestContractManager:
    def test_iter(self, smaller_project):
        actual = list(iter(smaller_project.contracts))
        assert len(actual) > 0
        assert "Other" in actual

    def test_compile(self, smaller_project):
        path = smaller_project.sources.lookup("Other.json")
        actual = list(smaller_project.contracts._compile(path))
        assert len(actual) == 1
        assert actual[0].contract_type.name == "Other"

        # Show it can happen again.
        actual = list(smaller_project.contracts._compile(path))
        assert len(actual) == 1
        assert actual[0].contract_type.name == "Other"

    def test_values(self, small_temp_project):
        actual = [c.name for c in small_temp_project.contracts.values()]
        assert "Other" in actual
        example = small_temp_project.contracts["Other"]
        count = len(small_temp_project.contracts)

        # Delete a file and try again, as a test.
        file = small_temp_project.path / example.source_id
        assert file.is_file()
        file.unlink()

        new_count = len(small_temp_project.contracts)
        assert new_count == count - 1


class TestDeploymentManager:
    def test_track(self, project, owner):
        instance = project.ContractC.deploy(sender=owner)
        project.deployments.track(instance, allow_dev=True)
        contract_type = instance.contract_type
        found = False
        for deployment in project.deployments:
            if deployment.contract_type == f"{contract_type.source_id}:{contract_type.name}":
                found = True
                break

        assert found

    def test_instance_map(self, project, vyper_contract_instance):
        project.deployments.track(vyper_contract_instance, allow_dev=True)
        assert project.deployments.instance_map != {}

        bip122_chain_id = to_hex(project.provider.get_block(0).hash)
        expected_uri = f"blockchain://{bip122_chain_id[2:]}/block/"
        for key in project.deployments.instance_map.keys():
            if key.startswith(expected_uri):
                return

        assert False, "Failed to find expected URI"


def test_chdir(smaller_project):
    original_path = smaller_project.path
    with create_tempdir() as new_path:
        smaller_project.chdir(new_path)
        assert smaller_project.path == new_path

        # Undo.
        smaller_project.chdir(original_path)
        assert smaller_project.path == original_path

        # Show you can also use it in a context.
        with smaller_project.chdir(new_path):
            assert smaller_project.path == new_path

        # It should have automatically undone.
        assert smaller_project.path == original_path


def test_within_project_path():
    start_cwd = Path.cwd()
    with create_tempdir() as new_path:
        project = Project(new_path)
        assert Path.cwd() != new_path

        with project.within_project_path():
            assert Path.cwd() == project.path

    assert Path.cwd() == start_cwd



from importlib import import_module
from pathlib import Path
from shutil import copytree
from typing import Optional

import pytest

from ape import Project
from tests.conftest import ApeSubprocessRunner

from .test_plugins import ListResult
from .utils import NodeId, __project_names__, __projects_directory__, project_skipper


class IntegrationTestModule:
    """
    A test module in 'tests.integration.cli'.
    """

    def __init__(self, path: Path):
        self._path = path
        module = import_module(f"tests.integration.cli.{path.stem}")
        test_methods = [
            getattr(module, t)
            for t in dir(module)
            if t.startswith("test_") and hasattr(t, "__name__") and hasattr(t, "__module__")
        ]
        self.tests = [NodeId(t) for t in test_methods]

    def __iter__(self):
        return iter(self.tests)

    @property
    def name(self) -> str:
        """
        The name of the module.
        """
        return self._path.stem


# Loads the actual test modules / methods
integration_tests = [
    IntegrationTestModule(p)
    for p in Path(__file__).parent.iterdir()
    if p.suffix == ".py" and p.name.startswith("test_")
]


@pytest.hookimpl(trylast=True)
def pytest_collection_modifyitems(session, config, items):
    """
    Filter out tests marked to be skipped using ``skip_projects``
    and the ``skip_projects_except`` decorators.
    """
    modified_items = []
    for item in items:
        item_name_parts = item.name.split("[")
        item_name_parts = [p.strip("[]") for p in item_name_parts]

        module_full_name = getattr(item.module, "__name__", None)
        if not module_full_name:
            continue

        module_name = module_full_name.split(".")[-1]
        test_name = item_name_parts[0]

        # Handle if a parametrized test is on-top
        # of the project's parametrization.
        project_name = item_name_parts[-1]
        for proj_name in project_skipper:
            # Example: 'test_foo[project-name-fuzz-0]' matches 'project-name'
            if project_name.startswith(proj_name):
                project_name = proj_name
                break

        is_cli_integration_test = (
            len(item_name_parts) == 2 and "integration.cli" in module_full_name
        )

        if not is_cli_integration_test or not project_skipper.do_skip(
            project_name, module_name, test_name
        ):
            modified_items.append(item)

    items[:] = modified_items


@pytest.fixture(autouse=True, scope="session")
def project_dir_map():
    """
    Ensure only copying projects once to prevent `TooManyOpenFilesError`.
    """

    class ProjectDirCache:
        project_map: dict[str, Path] = {}

        def load(self, name: str) -> Path:
            base_path = Path(__file__).parent / "projects"
            if name in self.project_map:
                res = self.project_map[name]
                if res.is_dir():
                    # Already copied and still exists!
                    return res

            # Either re-copy or copy for the first time.
            project_source_dir = __projects_directory__ / name
            project_dest_dir = base_path / project_source_dir.name
            project_dest_dir.parent.mkdir(exist_ok=True, parents=True)

            if not project_dest_dir.is_dir():
                copytree(project_source_dir, project_dest_dir)

            self.project_map[name] = project_dest_dir
            return self.project_map[name]

    return ProjectDirCache()


@pytest.fixture(autouse=True, params=__project_names__)
def integ_project(request, project_dir_map):
    project_dir = project_dir_map.load(request.param)
    if not project_dir.is_dir():
        # Should not happen because of logic in fixture,
        # but just in case!
        pytest.fail("Setup failed - project dir not exists.")

    root_project = Project(project_dir)
    # Using tmp project so no .build folder get kept around.
    with root_project.isolate_in_tempdir(name=request.param) as tmp_project:
        assert tmp_project.path.is_dir(), "Setup failed - tmp-project dir not exists"
        yield tmp_project


@pytest.fixture(scope="session")
def ape_cli():
    from ape._cli import cli

    return cli


def assert_failure(result, expected_output):
    assert result.exit_code == 1
    assert result.exception is not None
    assert "ERROR" in result.output
    assert expected_output in result.output


@pytest.fixture
def clean_cache(integ_project):
    """
    Use this fixture to ensure a project
    does not have a cached compilation.
    """
    integ_project.clean()
    yield
    integ_project.clean()


@pytest.fixture(scope="session")
def ape_plugins_runner(config):
    """
    Use subprocess runner so can manipulate site packages and see results.
    """

    class PluginSubprocessRunner(ApeSubprocessRunner):
        def __init__(self):
            super().__init__("plugins", data_folder=config.DATA_FOLDER)

        def invoke_list(self, arguments: list | None = None):
            arguments = arguments or []
            result = self.invoke("list", *arguments)
            assert result.exit_code == 0, result.output
            return ListResult.parse_output(result.output)

    return PluginSubprocessRunner()



import atexit
import ctypes
import datetime
import logging
import platform
import shutil
import sys
import time
import warnings
from abc import abstractmethod
from collections.abc import Iterable, Iterator
from functools import cached_property
from logging import FileHandler, Formatter, Logger, getLogger
from pathlib import Path
from signal import SIGINT, SIGTERM, signal
from subprocess import DEVNULL, PIPE, Popen
from typing import TYPE_CHECKING, Any, ClassVar, cast

from eth_pydantic_types import HexBytes
from eth_utils import to_hex
from ethpm_types import MethodABI
from pydantic import Field, computed_field, field_serializer, model_validator

from ape.api.networks import NetworkAPI
from ape.api.query import BlockTransactionQuery
from ape.api.transactions import ReceiptAPI, TransactionAPI
from ape.exceptions import (
    APINotImplementedError,
    ContractLogicError,
    ProviderError,
    QueryEngineError,
    RPCTimeoutError,
    SubprocessError,
    SubprocessTimeoutError,
    VirtualMachineError,
)
from ape.logging import LogLevel, logger
from ape.types.basic import HexInt
from ape.utils.basemodel import BaseInterfaceModel, BaseModel, ManagerAccessMixin
from ape.utils.misc import (
    EMPTY_BYTES32,
    _create_raises_not_implemented_error,
    log_instead_of_fail,
    raises_not_implemented,
    to_int,
)
from ape.utils.process import JoinableQueue, spawn
from ape.utils.rpc import RPCHeaders

if TYPE_CHECKING:
    from ethpm_types.abi import EventABI

    from ape.api.accounts import TestAccountAPI
    from ape.api.config import PluginConfig
    from ape.api.trace import TraceAPI
    from ape.types.address import AddressType
    from ape.types.events import ContractLog, LogFilter
    from ape.types.vm import BlockID, ContractCode, SnapshotID


class BlockAPI(BaseInterfaceModel):
    """
    An abstract class representing a block and its attributes.
    """

    # NOTE: All fields in this class (and its subclasses) should not be `Optional`
    #       except the edge cases noted below

    num_transactions: HexInt = 0
    """
    The number of transactions in the block.
    """

    hash: Any | None = None  # NOTE: pending block does not have a hash
    """
    The block hash identifier.
    """

    number: HexInt | None = None  # NOTE: pending block does not have a number
    """
    The block number identifier.
    """

    parent_hash: Any = Field(
        default=EMPTY_BYTES32, alias="parentHash"
    )  # NOTE: genesis block has no parent hash
    """
    The preceding block's hash.
    """

    timestamp: HexInt
    """
    The timestamp the block was produced.
    NOTE: The pending block uses the current timestamp.
    """

    _size: HexInt | None = None

    @log_instead_of_fail(default="<BlockAPI>")
    def __repr__(self) -> str:
        repr_str = f"{self.__class__.__name__} number={self.number}"
        if hash := self.hash:
            repr_str = f"{repr_str} hash={to_hex(hash)}"

        return f"<{repr_str}>"

    @property
    def datetime(self) -> datetime.datetime:
        """
        The block timestamp as a datetime object.
        """
        return datetime.datetime.fromtimestamp(self.timestamp, tz=datetime.timezone.utc)

    @model_validator(mode="before")
    @classmethod
    def convert_parent_hash(cls, data):
        parent_hash = data.get("parent_hash", data.get("parentHash")) or EMPTY_BYTES32
        data["parentHash"] = parent_hash
        return data

    @model_validator(mode="wrap")
    @classmethod
    def validate_size(cls, values, handler):
        """
        A validator for handling non-computed size.
        Saves it to a private member on this class and
        gets returned in computed field "size".
        """
        if isinstance(values, BlockAPI):
            size = values.size

        else:
            if not hasattr(values, "pop"):
                # Handle weird AttributeDict missing pop method.
                # https://github.com/ethereum/web3.py/issues/3326
                values = {**values}

            size = values.pop("size", None)

        model = handler(values)
        if size is not None:
            model._size = to_int(size)

        return model

    @field_serializer("size")
    def serialize_size(self, value):
        return to_int(value)

    @computed_field()  # type: ignore[misc]
    @cached_property
    def transactions(self) -> list[TransactionAPI]:
        """
        All transactions in a block.
        """
        if self.hash is None:
            # Unable to query transactions.
            return []

        try:
            query = BlockTransactionQuery(columns=["*"], block_id=self.hash)
            return cast(list[TransactionAPI], list(self.query_manager.query(query)))
        except QueryEngineError as err:
            # NOTE: Re-raising a better error here because was confusing
            #  when doing anything with fields, and this would fail.
            raise ProviderError(f"Unable to find block transactions: {err}") from err

    @computed_field()  # type: ignore[misc]
    @cached_property
    def size(self) -> HexInt:
        """
        The size of the block in gas. Most of the time,
        this field is passed to the model at validation time,
        but occasionally it is missing (like in `eth_subscribe:newHeads`),
        in which case it gets calculated if and only if the user
        requests it (or during serialization of this model to disk).
        """
        if self._size is not None:
            # The size was provided with the rest of the model
            # (normal).
            return self._size

        raise APINotImplementedError()


class CallResult(BaseModel, ManagerAccessMixin):
    """
    The result of a call.
    NOTE: Currently, you only get this for reverted calls from ``ProviderAPI``.
    """

    revert: ContractLogicError | None = None
    """
    The revert, if the call reverted.
    """

    returndata: HexBytes
    """
    The raw returndata.
    """

    @classmethod
    def from_revert(cls, revert: ContractLogicError) -> "CallResult":
        """
        Factory helper for easily creating a CallResult from a revert error.

        Args:
            revert (ContractLogicError): The revert error.

        Returns:
            A new ``CallResult`` instance.
        """
        return cls(revert=revert, returndata=HexBytes(revert.revert_message.encode("utf8").hex()))

    @property
    def reverted(self) -> bool:
        """
        ``True`` if the call `reverted` (failed).
        """
        return self.revert is not None

    @property
    def revert_message(self) -> str | None:
        """
        The revert message, if the call reverted.
        """
        return self.revert.revert_message if self.revert else None

    def decode(self, method_abi: MethodABI) -> Any:
        """
        Decode a Pythonic return value from the call.

        Args:
            method_abi (``MethodABI``): The method ABI to decode.

        Returns:
            The decoded result.
        """
        return self.provider.network.ecosystem.decode_returndata(method_abi, self.returndata)


class ProviderAPI(BaseInterfaceModel):
    """
    An abstraction of a connection to a network in an ecosystem. Example ``ProviderAPI``
    implementations include the `ape-infura <https://github.com/ApeWorX/ape-infura>`__
    plugin or the `ape-hardhat <https://github.com/ApeWorX/ape-hardhat>`__ plugin.
    """

    # TODO: In 0.9, make not optional.
    NAME: ClassVar[str | None] = None

    # TODO: Remove in 0.9 and have NAME be defined at the class-level (in plugins).
    name: str
    """(deprecated: use NAME). The name of the provider (should be the plugin name)."""

    network: NetworkAPI
    """A reference to the network this provider provides."""

    provider_settings: dict = {}
    """The settings for the provider, as overrides to the configuration."""

    # TODO: In 0.9, make @property that returns value from config,
    #   and use REQUEST_HEADER as plugin-defined constants.
    request_header: dict = {}
    """A header to set on HTTP/RPC requests."""

    block_page_size: int = 100
    """
    The amount of blocks to fetch in a response, as a default.
    This is particularly useful for querying logs across a block range.
    """

    concurrency: int = 4
    """
    How many parallel threads to use when fetching logs.
    """

    @property
    def data_folder(self) -> Path:
        """
        The path to the provider's data,
        e.g. ``$HOME/.api/{self.name}`` unless overridden.
        """
        return self.config_manager.DATA_FOLDER / self.name

    @property
    @abstractmethod
    def is_connected(self) -> bool:
        """
        ``True`` if currently connected to the provider. ``False`` otherwise.
        """

    @property
    def connection_str(self) -> str:
        """
        The str representing how to connect
        to the node, such as an HTTP URL
        or an IPC path.
        """
        return ""

    @abstractmethod
    def connect(self):
        """
        Connect to a provider, such as start-up a process or create an HTTP connection.
        """

    @abstractmethod
    def disconnect(self):
        """
        Disconnect from a provider, such as tear-down a process or quit an HTTP session.
        """

    @property
    def ipc_path(self) -> Path | None:
        """
        Return the IPC path for the provider, if supported.
        """
        return None

    @property
    def http_uri(self) -> str | None:
        """
        Return the raw HTTP/HTTPS URI to connect to this provider, if supported.
        """
        return None

    @property
    def ws_uri(self) -> str | None:
        """
        Return the raw WS/WSS URI to connect to this provider, if supported.
        """
        return None

    @property
    def settings(self) -> "PluginConfig":
        """
        The combination of settings from ``ape-config.yaml`` and ``.provider_settings``.
        """
        CustomConfig = self.config.__class__
        data = {**self.config.model_dump(), **self.provider_settings}
        return CustomConfig.model_validate(data)

    @property
    def connection_id(self) -> str | None:
        """
        A connection ID to uniquely identify and manage multiple
        connections to providers, especially when working with multiple
        providers of the same type, like multiple Geth --dev nodes.
        """

        try:
            chain_id = self.chain_id
        except Exception:
            if chain_id := self.settings.get("chain_id"):
                pass

            else:
                # A connection is required to obtain a chain ID for this provider.
                return None

        # NOTE: If other provider settings are different, ``.update_settings()``
        #    should be called.
        return f"{self.network_choice}:{chain_id}"

    @abstractmethod
    def update_settings(self, new_settings: dict):
        """
        Change a provider's setting, such as configure a new port to run on.
        May require a reconnect.

        Args:
            new_settings (dict): The new provider settings.
        """

    @property
    @abstractmethod
    def chain_id(self) -> int:
        """
        The blockchain ID.
        See `ChainList <https://chainlist.org/>`__ for a comprehensive list of IDs.
        """

    @abstractmethod
    def get_balance(self, address: "AddressType", block_id: "BlockID | None" = None) -> int:
        """
        Get the balance of an account.

        Args:
            address (:class:`~ape.types.address.AddressType`): The address of the account.
            block_id (:class:`~ape.types.BlockID`): Optionally specify a block
              ID. Defaults to using the latest block.

        Returns:
            int: The account balance.
        """

    @abstractmethod
    def get_code(self, address: "AddressType", block_id: "BlockID | None" = None) -> "ContractCode":
        """
        Get the bytes a contract.

        Args:
            address (:class:`~ape.types.address.AddressType`): The address of the contract.
            block_id ("BlockID | None"): The block ID
                  for checking a previous account nonce.

        Returns:
            :class:`~ape.types.ContractCode`: The contract bytecode.
        """

    @property
    def network_choice(self) -> str:
        """
        The connected network choice string.
        """
        if self.network.is_adhoc and self.connection_str:
            # `custom` is not a real network and is same
            # as using raw connection str
            return self.connection_str

        elif self.network.is_adhoc:
            raise ProviderError("Custom network provider missing `connection_str`.")

        return f"{self.network.choice}:{self.name}"

    @abstractmethod
    def make_request(self, rpc: str, parameters: Iterable | None = None) -> Any:
        """
        Make a raw RPC request to the provider.
        Advanced features such as tracing may utilize this to by-pass unnecessary
        class-serializations.
        """

    @raises_not_implemented
    def stream_request(  # type: ignore[empty-body]
        self, method: str, params: Iterable, iter_path: str = "result.item"
    ) -> Iterator[Any]:
        """
        Stream a request, great for large requests like events or traces.

        Args:
            method (str): The RPC method to call.
            params (Iterable): Parameters for the method.s
            iter_path (str): The response dict-path to the items.

        Returns:
            An iterator of items.
        """

    # TODO: In 0.9, delete this method.
    def get_storage_at(self, *args, **kwargs) -> HexBytes:
        warnings.warn(
            "'provider.get_storage_at()' is deprecated. Use 'provider.get_storage()'.",
            DeprecationWarning,
        )
        return self.get_storage(*args, **kwargs)

    @raises_not_implemented
    def get_storage(  # type: ignore[empty-body]
        self, address: "AddressType", slot: int, block_id: "BlockID | None" = None
    ) -> HexBytes:
        """
        Gets the raw value of a storage slot of a contract.

        Args:
            address (AddressType): The address of the contract.
            slot (int): Storage slot to read the value of.
            block_id ("BlockID | None"): The block ID
              for checking a previous storage value.

        Returns:
            HexBytes: The value of the storage slot.
        """

    @abstractmethod
    def get_nonce(self, address: "AddressType", block_id: "BlockID | None" = None) -> int:
        """
        Get the number of times an account has transacted.

        Args:
            address (AddressType): The address of the account.
            block_id ("BlockID | None"): The block ID
              for checking a previous account nonce.

        Returns:
            int
        """

    @abstractmethod
    def estimate_gas_cost(self, txn: TransactionAPI, block_id: "BlockID | None" = None) -> int:
        """
        Estimate the cost of gas for a transaction.

        Args:
            txn (:class:`~ape.api.transactions.TransactionAPI`):
              The transaction to estimate the gas for.
            block_id ("BlockID | None"): The block ID
              to use when estimating the transaction. Useful for checking a
              past estimation cost of a transaction.

        Returns:
            int: The estimated cost of gas to execute the transaction
            reported in the fee-currency's smallest unit, e.g. Wei. If the
            provider's network has been configured with a gas limit override, it
            will be returned. If the gas limit configuration is "max" this will
            return the block maximum gas limit.
        """

    @property
    @abstractmethod
    def gas_price(self) -> int:
        """
        The price for what it costs to transact
        (pre-`EIP-1559 <https://eips.ethereum.org/EIPS/eip-1559>`__).
        """

    @property
    @abstractmethod
    def max_gas(self) -> int:
        """
        The max gas limit value you can use.
        """

    @property
    def config(self) -> "PluginConfig":
        """
        The provider's configuration.
        """
        return self.config_manager.get_config(self.name)

    @property
    def priority_fee(self) -> int:
        """
        A miner tip to incentivize them to include your transaction in a block.

        Raises:
            NotImplementedError: When the provider does not implement
              `EIP-1559 <https://eips.ethereum.org/EIPS/eip-1559>`__ typed transactions.
        """
        raise APINotImplementedError("priority_fee is not implemented by this provider")

    @property
    def supports_tracing(self) -> bool:
        """
        ``True`` when the provider can provide transaction traces.
        """
        return False

    @property
    def base_fee(self) -> int:
        """
        The minimum value required to get your transaction included on the next block.
        Only providers that implement `EIP-1559 <https://eips.ethereum.org/EIPS/eip-1559>`__
        will use this property.

        Raises:
            NotImplementedError: When this provider does not implement
              `EIP-1559 <https://eips.ethereum.org/EIPS/eip-1559>`__.
        """
        raise APINotImplementedError("base_fee is not implemented by this provider")

    @abstractmethod
    def get_block(self, block_id: "BlockID") -> BlockAPI:
        """
        Get a block.

        Args:
            block_id (:class:`~ape.types.BlockID`): The ID of the block to get.
                Can be ``"latest"``, ``"earliest"``, ``"pending"``, a block hash or a block number.

        Raises:
            :class:`~ape.exceptions.BlockNotFoundError`: Likely the exception raised when a block
              is not found (depends on implementation).

        Returns:
            :class:`~ape.types.BlockID`: The block for the given ID.
        """

    # TODO: In 0.9, change the return value to be `CallResult`
    #    (right now it does only when using raise_on_revert=False and it reverts).
    @abstractmethod
    def send_call(
        self,
        txn: TransactionAPI,
        block_id: "BlockID | None" = None,
        state: dict | None = None,
        **kwargs,
    ) -> HexBytes:  # Return value of function
        """
        Execute a new transaction call immediately without creating a
        transaction on the blockchain.

        Args:
            txn: :class:`~ape.api.transactions.TransactionAPI`
            block_id ("BlockID | None"): The block ID
                to use to send a call at a historical point of a contract.
                Useful for checking a past estimation cost of a transaction.
            state (Optional[dict]): Modify the state of the blockchain
                prior to sending the call, for testing purposes.
            **kwargs: Provider-specific extra kwargs.

        Returns:
            HexBytes: The returndata of the transaction call. Even though it isn't
            mentioned in the type, you can also return a ``:class:`~ape.api.providers.CallResult``
            here and Ape will handle it. In 0.9, it will be the primary return type.
        """

    @abstractmethod
    def get_receipt(self, txn_hash: str, **kwargs) -> ReceiptAPI:
        """
        Get the information about a transaction from a transaction hash.

        Args:
            txn_hash (str): The hash of the transaction to retrieve.
            kwargs: Any other kwargs that other providers might allow when fetching a receipt.

        Returns:
            :class:`~api.providers.ReceiptAPI`:
            The receipt of the transaction with the given hash.
        """

    @abstractmethod
    def get_transactions_by_block(self, block_id: "BlockID") -> Iterator[TransactionAPI]:
        """
        Get the information about a set of transactions from a block.

        Args:
            block_id (:class:`~ape.types.BlockID`): The ID of the block.

        Returns:
            Iterator[:class: `~ape.api.transactions.TransactionAPI`]
        """

    @raises_not_implemented
    def get_transactions_by_account_nonce(  # type: ignore[empty-body]
        self,
        account: "AddressType",
        start_nonce: int = 0,
        stop_nonce: int = -1,
    ) -> Iterator[ReceiptAPI]:
        """
        Get account history for the given account.

        Args:
            account (:class:`~ape.types.address.AddressType`): The address of the account.
            start_nonce (int): The nonce of the account to start the search with.
            stop_nonce (int): The nonce of the account to stop the search with.

        Returns:
            Iterator[:class:`~ape.api.transactions.ReceiptAPI`]
        """

    @abstractmethod
    def send_transaction(self, txn: TransactionAPI) -> ReceiptAPI:
        """
        Send a transaction to the network.

        Args:
            txn (:class:`~ape.api.transactions.TransactionAPI`): The transaction to send.

        Returns:
            :class:`~ape.api.transactions.ReceiptAPI`
        """

    @abstractmethod
    def get_contract_logs(self, log_filter: "LogFilter") -> Iterator["ContractLog"]:
        """
        Get logs from contracts.

        Args:
            log_filter (:class:`~ape.types.LogFilter`): A mapping of event ABIs to
              topic filters. Defaults to getting all events.

        Returns:
            Iterator[:class:`~ape.types.ContractLog`]
        """

    def send_private_transaction(self, txn: TransactionAPI, **kwargs) -> ReceiptAPI:
        """
        Send a transaction through a private mempool (if supported by the Provider).

        Raises:
            :class:`~ape.exceptions.APINotImplementedError`: If using a non-local
              network and not implemented by the provider.

        Args:
            txn (:class:`~ape.api.transactions.TransactionAPI`): The transaction
              to privately publish.
            **kwargs: Additional kwargs to be optionally handled by the provider.

        Returns:
            :class:`~ape.api.transactions.ReceiptAPI`
        """
        if self.network.is_dev:
            # Send the transaction as normal so testers can verify private=True
            # and the txn still goes through.
            logger.warning(
                f"private=True is set but connected to network '{self.network.name}' ."
                f"Using regular '{self.send_transaction.__name__}()' method (not private)."
            )
            return self.send_transaction(txn)

        # What happens normally from `raises_not_implemented()` decorator.
        raise _create_raises_not_implemented_error(self.send_private_transaction)

    @raises_not_implemented
    def snapshot(self) -> "SnapshotID":  # type: ignore[empty-body]
        """
        Defined to make the ``ProviderAPI`` interchangeable with a
        :class:`~ape.api.providers.TestProviderAPI`, as in
        :class:`ape.managers.chain.ChainManager`.

        Raises:
            :class:`~ape.exceptions.APINotImplementedError`: Unless overridden.
        """

    @raises_not_implemented
    def restore(self, snapshot_id: "SnapshotID"):
        """
        Defined to make the ``ProviderAPI`` interchangeable with a
        :class:`~ape.api.providers.TestProviderAPI`, as in
        :class:`ape.managers.chain.ChainManager`.

        Raises:
            :class:`~ape.exceptions.APINotImplementedError`: Unless overridden.
        """

    @raises_not_implemented
    def set_timestamp(self, new_timestamp: int):
        """
        Defined to make the ``ProviderAPI`` interchangeable with a
        :class:`~ape.api.providers.TestProviderAPI`, as in
        :class:`ape.managers.chain.ChainManager`.

        Raises:
            :class:`~ape.exceptions.APINotImplementedError`: Unless overridden.
        """

    @raises_not_implemented
    def mine(self, num_blocks: int = 1):
        """
        Defined to make the ``ProviderAPI`` interchangeable with a
        :class:`~ape.api.providers.TestProviderAPI`, as in
        :class:`ape.managers.chain.ChainManager`.

        Raises:
            :class:`~ape.exceptions.APINotImplementedError`: Unless overridden.
        """

    @raises_not_implemented
    def set_balance(self, address: "AddressType", amount: int):
        """
        Change the balance of an account.

        Args:
            address (AddressType): An address on the network.
            amount (int): The balance to set in the address.
        """

    @raises_not_implemented
    def get_test_account(self, index: int) -> "TestAccountAPI":  # type: ignore[empty-body]
        """
        Retrieve one of the provider-generated test accounts.

        Args:
            index (int): The index of the test account in the HD-Path.

        Returns:
            :class:`~ape.api.accounts.TestAccountAPI`
        """

    @log_instead_of_fail(default="<ProviderAPI>")
    def __repr__(self) -> str:
        return f"<{self.name.capitalize()} chain_id={self.chain_id}>"

    @raises_not_implemented
    def set_code(  # type: ignore[empty-body]
        self, address: "AddressType", code: "ContractCode"
    ) -> bool:
        """
        Change the code of a smart contract, for development purposes.
        Test providers implement this method when they support it.

        Args:
            address (AddressType): An address on the network.
            code (:class:`~ape.types.ContractCode`): The new bytecode.
        """

    @raises_not_implemented
    def set_storage(  # type: ignore[empty-body]
        self, address: "AddressType", slot: int, value: HexBytes
    ):
        """
        Sets the raw value of a storage slot of a contract.

        Args:
            address (str): The address of the contract.
            slot (int): Storage slot to write the value to.
            value: (HexBytes): The value to overwrite the raw storage slot with.
        """

    @raises_not_implemented
    def unlock_account(self, address: "AddressType") -> bool:  # type: ignore[empty-body]
        """
        Ask the provider to allow an address to submit transactions without validating
        signatures. This feature is intended to be subclassed by a
        :class:`~ape.api.providers.TestProviderAPI` so that during a fork-mode test,
        a transaction can be submitted by an arbitrary account or contract without a private key.

        Raises:
            NotImplementedError: When this provider does not support unlocking an account.

        Args:
            address (:class:`~ape.types.address.AddressType`): The address to unlock.

        Returns:
            bool: ``True`` if successfully unlocked account and ``False`` otherwise.
        """

    @raises_not_implemented
    def relock_account(self, address: "AddressType"):
        """
        Stop impersonating an account.

        Args:
            address (:class:`~ape.types.address.AddressType`): The address to relock.
        """

    @raises_not_implemented
    def get_transaction_trace(  # type: ignore[empty-body]
        self, txn_hash: HexBytes | str
    ) -> "TraceAPI":
        """
        Provide a detailed description of opcodes.

        Args:
            txn_hash (Union[HexBytes, str]): The hash of a transaction
              to trace.

        Returns:
            :class:`~ape.api.trace.TraceAPI`: A transaction trace.
        """

    @raises_not_implemented
    def poll_blocks(  # type: ignore[empty-body]
        self,
        stop_block: int | None = None,
        required_confirmations: int | None = None,
        new_block_timeout: int | None = None,
    ) -> Iterator[BlockAPI]:
        """
        Poll new blocks.

        **NOTE**: When a chain reorganization occurs, this method logs an error and
        yields the missed blocks, even if they were previously yielded with different
        block numbers.

        **NOTE**: This is a daemon method; it does not terminate unless an exception occurs
        or a ``stop_block`` is given.

        Args:
            stop_block (Optional[int]): Optionally set a future block number to stop at.
              Defaults to never-ending.
            required_confirmations (Optional[int]): The amount of confirmations to wait
              before yielding the block. The more confirmations, the less likely a reorg will occur.
              Defaults to the network's configured required confirmations.
            new_block_timeout (Optional[float]): The amount of time to wait for a new block before
              timing out. Defaults to 10 seconds for local networks or ``50 * block_time`` for live
              networks.

        Returns:
            Iterator[:class:`~ape.api.providers.BlockAPI`]
        """

    @raises_not_implemented
    def poll_logs(  # type: ignore[empty-body]
        self,
        stop_block: int | None = None,
        address: "AddressType | None" = None,
        topics: list[str | list[str]] | None = None,
        required_confirmations: int | None = None,
        new_block_timeout: int | None = None,
        events: list["EventABI"] | None = None,
    ) -> Iterator["ContractLog"]:
        """
        Poll new blocks. Optionally set a start block to include historical blocks.

        **NOTE**: This is a daemon method; it does not terminate unless an exception occurs.

        Usage example::

            for new_log in contract.MyEvent.poll_logs():
                print(f"New event log found: block_number={new_log.block_number}")

        Args:
            stop_block (Optional[int]): Optionally set a future block number to stop at.
              Defaults to never-ending.
            address (Optional[str]): The address of the contract to filter logs by.
              Defaults to all addresses.
            topics (Optional[list[Union[str, list[str]]]]): The topics to filter logs by.
              Defaults to all topics.
            required_confirmations (Optional[int]): The amount of confirmations to wait
              before yielding the block. The more confirmations, the less likely a reorg will occur.
              Defaults to the network's configured required confirmations.
            new_block_timeout (Optional[int]): The amount of time to wait for a new block before
              quitting. Defaults to 10 seconds for local networks or ``50 * block_time`` for live
              networks.
            events (Optional[list[``EventABI``]]): An optional list of events to listen on.

        Returns:
            Iterator[:class:`~ape.types.ContractLog`]
        """

    def prepare_transaction(self, txn: TransactionAPI) -> TransactionAPI:
        """
        Set default values on the transaction.

        Raises:
            :class:`~ape.exceptions.TransactionError`: When given negative required confirmations.

        Args:
            txn (:class:`~ape.api.transactions.TransactionAPI`): The transaction to prepare.

        Returns:
            :class:`~ape.api.transactions.TransactionAPI`
        """
        return txn

    def get_virtual_machine_error(self, exception: Exception, **kwargs) -> VirtualMachineError:
        """
        Get a virtual machine error from an error returned from your RPC.

        Args:
            exception (Exception): The error returned from your RPC client.

        Returns:
            :class:`~ape.exceptions.VirtualMachineError`: An error representing what
               went wrong in the call.
        """
        return VirtualMachineError(base_err=exception, **kwargs)

    def _get_request_headers(self) -> RPCHeaders:
        # Internal helper method called by NetworkManager
        headers = RPCHeaders(**self.request_header)
        # Have to do it this way to avoid "multiple-keys" error.
        configured_headers: dict = self.config.get("request_headers", {})
        for key, value in configured_headers.items():
            headers[key] = value

        return headers


class TestProviderAPI(ProviderAPI):
    """
    An API for providers that have development functionality, such as snapshotting.
    """

    @cached_property
    def test_config(self) -> "PluginConfig":
        return self.config_manager.get_config("test")

    @abstractmethod
    def snapshot(self) -> "SnapshotID":
        """
        Record the current state of the blockchain with intent to later
        call the method :meth:`~ape.managers.chain.ChainManager.revert`
        to go back to this point. This method is for local networks only.

        Returns:
            :class:`~ape.types.SnapshotID`: The snapshot ID.
        """

    @abstractmethod
    def restore(self, snapshot_id: "SnapshotID"):
        """
        Regress the current call using the given snapshot ID.
        Allows developers to go back to a previous state.

        Args:
            snapshot_id (str): The snapshot ID.
        """

    @abstractmethod
    def set_timestamp(self, new_timestamp: int):
        """
        Change the pending timestamp.

        Args:
            new_timestamp (int): The timestamp to set.

        Returns:
            int: The new timestamp.
        """

    @abstractmethod
    def mine(self, num_blocks: int = 1):
        """
        Advance by the given number of blocks.

        Args:
            num_blocks (int): The number of blocks allotted to mine. Defaults to ``1``.
        """

    @property
    @abstractmethod
    def auto_mine(self) -> bool:
        """
        Whether automine is enabled.
        """

    @auto_mine.setter
    @abstractmethod
    def auto_mine(self, value) -> bool:
        """
        Enable or disable automine.
        """

    def _increment_call_func_coverage_hit_count(self, txn: TransactionAPI):
        """
        A helper method for incrementing a method call function hit count in a
        non-orthodox way. This is because Hardhat does not support call traces yet.
        """
        if (
            not txn.receiver
            or not self._test_runner
            or not self._test_runner.config_wrapper.track_coverage
        ):
            return

        if not (contract_type := self.chain_manager.contracts.get(txn.receiver)) or not (
            contract_src := self.local_project._create_contract_source(contract_type)
        ):
            return

        method_id = txn.data[:4]
        if method_id in contract_type.view_methods:
            method = contract_type.methods[method_id]
            self._test_runner.coverage_tracker.hit_function(contract_src, method)


class UpstreamProvider(ProviderAPI):
    """
    A provider that can also be set as another provider's upstream.
    """


class SubprocessProvider(ProviderAPI):
    """
    A provider that manages a process, such as for ``ganache``.
    """

    PROCESS_WAIT_TIMEOUT: int = 15
    background: bool = False
    process: Popen | None = None
    allow_start: bool = True
    is_stopping: bool = False

    stdout_queue: JoinableQueue | None = None
    stderr_queue: JoinableQueue | None = None

    @property
    @abstractmethod
    def process_name(self) -> str:
        """The name of the process, such as ``Hardhat node``."""

    @abstractmethod
    def build_command(self) -> list[str]:
        """
        Get the command as a list of ``str``.
        Subclasses should override and add command arguments if needed.

        Returns:
            list[str]: The command to pass to ``subprocess.Popen``.
        """

    @property
    def base_logs_path(self) -> Path:
        return self.config_manager.DATA_FOLDER / self.name / "subprocess_output"

    @property
    def stdout_logs_path(self) -> Path:
        return self.base_logs_path / "stdout.log"

    @property
    def stderr_logs_path(self) -> Path:
        return self.base_logs_path / "stderr.log"

    @cached_property
    def _stdout_logger(self) -> Logger:
        return self._get_process_output_logger("stdout", self.stdout_logs_path)

    @cached_property
    def _stderr_logger(self) -> Logger:
        return self._get_process_output_logger("stderr", self.stderr_logs_path)

    @property
    def connection_id(self) -> str | None:
        cmd_id = ",".join(self.build_command())
        return f"{self.network_choice}:{cmd_id}"

    def _get_process_output_logger(self, name: str, path: Path):
        logger = getLogger(f"{self.name}_{name}_subprocessProviderLogger")
        path.parent.mkdir(parents=True, exist_ok=True)
        if path.is_file():
            path.unlink()

        path.touch()
        handler = FileHandler(str(path))
        handler.setFormatter(Formatter("%(message)s"))
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)
        return logger

    def connect(self):
        """
        Start the process and connect to it.
        Subclasses handle the connection-related tasks.
        """

        if self.is_connected:
            raise ProviderError("Cannot connect twice. Call disconnect before connecting again.")

        # Always disconnect after,
        # unless running tests with `disconnect_providers_after: false`.
        disconnect_after = (
            self._test_runner is None
            or self.config_manager.get_config("test").disconnect_providers_after
        )
        if disconnect_after:
            atexit.register(self._disconnect_atexit)

        # Register handlers to ensure atexit handlers are called when Python dies.
        def _signal_handler(signum, frame):
            atexit._run_exitfuncs()
            sys.exit(143 if signum == SIGTERM else 130)

        signal(SIGINT, _signal_handler)
        signal(SIGTERM, _signal_handler)

    def _disconnect_atexit(self):
        if self.background:
            return

        try:
            self.disconnect()
        except Exception as err:
            logger.error(f"Error while disconnecting: {err}")

    def disconnect(self):
        """
        Stop the process if it exists.
        Subclasses override this method to do provider-specific disconnection tasks.
        """

        if self.process:
            self.stop()

        # Delete entry from managed list of running nodes.
        self.network_manager.running_nodes.remove_provider(self)

    def start(self, timeout: int = 20):
        """Start the process and wait for its RPC to be ready."""

        if self.is_connected:
            logger.info(f"Connecting to existing '{self.process_name}' process.")
            self.process = None  # Not managing the process.

        elif self.allow_start:
            logger.info(f"Starting '{self.process_name}' process.")
            pre_exec_fn = _linux_set_death_signal if platform.uname().system == "Linux" else None
            self.stderr_queue = JoinableQueue()
            self.stdout_queue = JoinableQueue()

            if self.background or logger.level > LogLevel.DEBUG:
                out_file = DEVNULL
            else:
                out_file = PIPE

            cmd = self.build_command()
            process = popen(cmd, preexec_fn=pre_exec_fn, stdout=out_file, stderr=out_file)
            self.process = process
            spawn(self.produce_stdout_queue)
            spawn(self.produce_stderr_queue)
            spawn(self.consume_stdout_queue)
            spawn(self.consume_stderr_queue)

            # Cache the process so we can manage it even if lost.
            self.network_manager.running_nodes.cache_provider(self)

            with RPCTimeoutError(self, seconds=timeout) as _timeout:
                while True:
                    if self.is_connected:
                        break

                    time.sleep(0.1)
                    _timeout.check()

        else:
            raise ProviderError("Process not started and cannot connect to existing process.")

    def produce_stdout_queue(self):
        process = self.process
        if self.stdout_queue is None or process is None:
            return

        stdout = process.stdout
        if stdout is None:
            return

        for line in iter(stdout.readline, b""):
            self.stdout_queue.put(line)
            time.sleep(0)

    def produce_stderr_queue(self):
        process = self.process
        if self.stderr_queue is None or process is None:
            return

        stderr = process.stderr
        if stderr is None:
            return

        for line in iter(stderr.readline, b""):
            self.stderr_queue.put(line)
            time.sleep(0)

    def consume_stdout_queue(self):
        if self.stdout_queue is None:
            return

        for line in self.stdout_queue:
            output = line.decode("utf8").strip()
            logger.debug(output)
            self._stdout_logger.debug(output)

            if self.stdout_queue is not None:
                self.stdout_queue.task_done()

            time.sleep(0)

    def consume_stderr_queue(self):
        if self.stderr_queue is None:
            return

        for line in self.stderr_queue:
            logger.debug(line.decode("utf8").strip())
            self._stdout_logger.debug(line)

            if self.stderr_queue is not None:
                self.stderr_queue.task_done()

            time.sleep(0)

    def stop(self):
        """Kill the process."""

        if not self.process or self.is_stopping:
            return

        self.is_stopping = True
        logger.info(f"Stopping '{self.process_name}' process.")
        self._kill_process()
        self.is_stopping = False
        self.process = None

    def _wait_for_popen(self, timeout: int = 30):
        if not self.process:
            # Mostly just to make mypy happy.
            raise SubprocessError("Unable to wait for process. It is not set yet.")

        try:
            with SubprocessTimeoutError(self, seconds=timeout) as _timeout:
                while self.process.poll() is None:
                    time.sleep(0.1)
                    _timeout.check()

        except SubprocessTimeoutError:
            pass

    def _kill_process(self):
        if platform.uname().system == "Windows":
            self._windows_taskkill()
            return

        warn_prefix = f"Trying to close '{self.process_name}' process."

        def _try_close(warn_message):
            try:
                if self.process:
                    self.process.send_signal(SIGINT)

                self._wait_for_popen(self.PROCESS_WAIT_TIMEOUT)
            except KeyboardInterrupt:
                logger.warning(warn_message)

        try:
            if self.process is not None and self.process.poll() is None:
                _try_close(f"{warn_prefix}. Press Ctrl+C 1 more times to force quit")

            if self.process is not None and self.process.poll() is None:
                self.process.kill()
                self._wait_for_popen(2)

        except KeyboardInterrupt:
            if self.process is not None:
                self.process.kill()

        self.process = None

    def _windows_taskkill(self) -> None:
        """
        Kills the given process and all child processes using taskkill.exe. Used
        for subprocesses started up on Windows which run in a cmd.exe wrapper that
        doesn't propagate signals by default (leaving orphaned processes).
        """
        process = self.process
        if not process:
            return

        taskkill_bin = shutil.which("taskkill")
        if not taskkill_bin:
            raise SubprocessError("Could not find taskkill.exe executable.")

        proc = Popen(
            [
                taskkill_bin,
                "/F",  # forcefully terminate
                "/T",  # terminate child processes
                "/PID",
                str(process.pid),
            ]
        )
        proc.wait(timeout=self.PROCESS_WAIT_TIMEOUT)


def _linux_set_death_signal():
    """
    Automatically sends SIGTERM to child subprocesses when parent process
    dies (only usable on Linux).
    """
    # from: https://stackoverflow.com/a/43152455/75956
    # the first argument, 1, is the flag for PR_SET_PDEATHSIG
    # the second argument is what signal to send to child subprocesses
    libc = ctypes.CDLL("libc.so.6")
    return libc.prctl(1, SIGTERM)


def popen(cmd: list[str], **kwargs):
    # Abstracted for testing purposes.
    return Popen(cmd, **kwargs)



import copy
from pathlib import Path
from typing import ClassVar
from unittest import mock

import pytest

from ape.api.config import PluginConfig
from ape.api.networks import ForkedNetworkAPI, NetworkAPI, create_network_type
from ape.api.providers import ProviderAPI
from ape.exceptions import NetworkError, ProviderNotFoundError
from ape.types.gas import AutoGasLimit
from ape_ethereum import Ethereum, EthereumConfig
from ape_ethereum.ecosystem import BaseEthereumConfig, NetworkConfig, create_network_config
from ape_ethereum.transactions import TransactionType


def test_get_provider_when_not_found(ethereum):
    name = "sepolia-fork"
    network = ethereum.get_network(name)
    expected = f"No provider named 'test' in network '{name}' in ecosystem 'ethereum'.*"
    with pytest.raises(ProviderNotFoundError, match=expected):
        network.get_provider("test")


@pytest.mark.parametrize("scheme", ("http", "https", "ws", "wss"))
def test_get_provider_http(ethereum, scheme):
    uri = f"{scheme}://example.com"
    network = ethereum.get_network("sepolia")
    actual = network.get_provider(uri)
    assert actual.uri == uri
    assert actual.network.name == "sepolia"


def test_get_provider_ipc(ethereum):
    path = "path/to/geth.ipc"
    network = ethereum.get_network("sepolia")
    actual = network.get_provider(path)
    assert actual.ipc_path == Path(path)
    assert actual.network.name == "sepolia"


def test_get_provider_custom_network(project, custom_networks_config_dict, ethereum):
    with project.temp_config(**custom_networks_config_dict):
        network = ethereum.apenet
        actual = network.get_provider("node")
        assert isinstance(actual, ProviderAPI)
        assert actual.name == "node"


def test_block_times(ethereum):
    assert ethereum.sepolia.block_time == 15


def test_set_default_provider_not_exists(ape_caplog, ethereum):
    bad_provider = "NOT_EXISTS"
    expected = f"Provider '{bad_provider}' not found in network 'ethereum:sepolia'."
    with pytest.raises(NetworkError, match=expected):
        ethereum.sepolia.set_default_provider(bad_provider)


def test_gas_limit_defaults(ethereum, project, custom_networks_config_dict):
    """
    Test the default gas limit configurations for local and live networks.
    """
    with project.temp_config(**custom_networks_config_dict):
        assert ethereum.sepolia.gas_limit == AutoGasLimit(multiplier=1.0)
        assert ethereum.apenet.gas_limit == AutoGasLimit(multiplier=1.0)
        assert ethereum.local.gas_limit == "max"


@pytest.mark.parametrize(
    "gas_limit_cfg",
    (
        {"auto": {"multiplier": 1.1}},
        AutoGasLimit(multiplier=1.1),
    ),
)
def test_gas_limit_configured_auto_with_multiplier(gas_limit_cfg):
    cfg = NetworkConfig(gas_limit=gas_limit_cfg)
    assert cfg.gas_limit == AutoGasLimit(multiplier=1.1)


def test_gas_limit_configured_auto(ethereum, project):
    network = NetworkConfig(gas_limit="auto")
    assert network.gas_limit == AutoGasLimit(multiplier=1.0)


def test_gas_limit_configured_max(ethereum, project):
    with project.temp_config(ethereum={"local": {"gas_limit": "max"}}):
        assert ethereum.local.gas_limit == "max"


@pytest.mark.parametrize(
    "gas_limit_cfg",
    (100, "0x64"),
)
def test_gas_limit_configured_numerically(gas_limit_cfg):
    network = NetworkConfig(gas_limit=gas_limit_cfg)
    assert network.gas_limit == 100


def test_base_fee_multiplier(ethereum):
    assert ethereum.mainnet.base_fee_multiplier == 1.4
    assert ethereum.local.base_fee_multiplier == 1.0


def test_forked_networks(ethereum):
    mainnet_fork = ethereum.mainnet_fork
    ethereum.mainnet._default_provider = "node"  # In case changed elsewhere in env.
    assert mainnet_fork.upstream_network.name == "mainnet"
    assert mainnet_fork.upstream_chain_id == 1
    # Just make sure it doesn't fail when trying to access.
    assert mainnet_fork.upstream_provider
    # Ensure has default configurations.
    cfg = mainnet_fork.ecosystem_config.mainnet_fork
    assert cfg.default_transaction_type == TransactionType.DYNAMIC
    assert cfg.block_time == 0
    assert cfg.default_provider is None
    assert cfg.base_fee_multiplier == 1.0
    assert cfg.transaction_acceptance_timeout == 20
    assert cfg.max_receipt_retries == 20


def test_forked_network_with_config(project, ethereum):
    data = {
        "ethereum": {"mainnet_fork": {"default_transaction_type": TransactionType.STATIC.value}}
    }
    with project.temp_config(**data):
        cfg = ethereum.mainnet_fork.config
        assert cfg.default_transaction_type == TransactionType.STATIC
        assert cfg.block_time == 0
        assert cfg.default_provider is None
        assert cfg.base_fee_multiplier == 1.0
        assert cfg.transaction_acceptance_timeout == 20
        assert cfg.max_receipt_retries == 20
        assert cfg.gas_limit == "max"


def test_data_folder_custom_network(custom_network, ethereum, custom_network_name_0):
    actual = custom_network.data_folder
    expected = ethereum.data_folder / custom_network_name_0
    assert actual == expected


def test_config_custom_networks_default(ethereum, project, custom_networks_config_dict):
    """
    Shows you don't get AttributeError when custom network config is not
    present.
    """
    with project.temp_config(**custom_networks_config_dict):
        network = ethereum.apenet
        cfg = network.config
        assert cfg.default_transaction_type == TransactionType.DYNAMIC


def test_config_custom_networks(
    ethereum, custom_networks_config_dict, project, custom_network_name_0
):
    data = copy.deepcopy(custom_networks_config_dict)
    data["ethereum"] = {
        custom_network_name_0: {"default_transaction_type": TransactionType.STATIC.value}
    }
    with project.temp_config(**data):
        network = ethereum.apenet
        ethereum_config = network.ecosystem_config
        cfg_by_attr = ethereum_config.apenet
        assert cfg_by_attr.default_transaction_type == TransactionType.STATIC

        assert "apenet" in ethereum_config
        cfg_by_get = ethereum_config.get("apenet")
        assert cfg_by_get is not None
        assert cfg_by_get.default_transaction_type == TransactionType.STATIC


def test_config_networks_from_custom_ecosystem(
    networks, custom_networks_config_dict, project, custom_network_name_0
):
    data = copy.deepcopy(custom_networks_config_dict)
    data["networks"]["custom"][0]["ecosystem"] = "custom-ecosystem"
    data["networks"]["custom"][1]["ecosystem"] = "custom-ecosystem"
    data["custom-ecosystem"] = {
        custom_network_name_0: {"default_transaction_type": TransactionType.STATIC.value}
    }
    with project.temp_config(**data):
        custom_ecosystem = networks.get_ecosystem("custom-ecosystem")
        network = custom_ecosystem.get_network("apenet")
        ecosystem_config = network.ecosystem_config
        network_by_attr = ecosystem_config.apenet
        network_by_get = ecosystem_config.get("apenet")

    assert custom_ecosystem.name == "custom-ecosystem"

    # Show .get_network() works (raises when not found).
    assert network.name == "apenet"

    # Show custom ecosystems have a config.
    assert isinstance(ecosystem_config, EthereumConfig)

    # Show contains works.
    assert "apenet" in ecosystem_config

    # Show dot-access works (raise AttrError when not found).
    assert network_by_attr.default_transaction_type == TransactionType.STATIC

    # Show .get() works (returns None when not found).
    assert network_by_get is not None
    assert network_by_get.default_transaction_type == TransactionType.STATIC


def test_config_validates_dict():
    """
    Ecosystem plugins may return dictionaries here.
    """
    val = 123
    name = "foodnetcfgtest"

    class MockEcosystem:
        @property
        def config(self):
            return {name: {"fooprop": val}}

    mock_ecosystem = MockEcosystem()
    network_type = create_network_type(0, 0, False)
    network = network_type.model_construct(name=name, ecosystem=mock_ecosystem)  # type: ignore

    assert isinstance(network.config, PluginConfig)
    assert network.config["fooprop"] == val
    assert network.config.fooprop == val  # getattr works (wouldn't if only a dict).


def test_use_provider_using_provider_instance(eth_tester_provider):
    network = eth_tester_provider.network
    with network.use_provider(eth_tester_provider) as provider:
        assert id(provider) == id(eth_tester_provider)


def test_use_provider_previously_used_and_not_connected(eth_tester_provider):
    network = eth_tester_provider.network
    eth_tester_provider.disconnect()
    with network.use_provider("test") as provider:
        assert provider.is_connected


def test_create_network_type():
    chain_id = 123321123321123321
    actual = create_network_type(chain_id, chain_id)
    assert issubclass(actual, NetworkAPI)


def test_create_network_type_fork():
    chain_id = 123321123321123322
    actual = create_network_type(chain_id, chain_id, is_fork=True)
    assert issubclass(actual, NetworkAPI)
    assert issubclass(actual, ForkedNetworkAPI)


def test_providers(ethereum):
    network = ethereum.local
    providers = network.providers
    assert "test" in providers
    assert "node" in providers


def test_providers_NAME_defined(mocker):
    """
    Show that when a provider plugin used the NAME ClassVar,
    the provider's name is that instead of the plugin's name.
    """
    ecosystem_name = "my-ecosystem"
    network_name = "my-network"
    provider_name = "my-provider"

    class MyProvider(ProviderAPI):
        NAME: ClassVar[str] = provider_name

    class MyNetwork(NetworkAPI):
        def _get_plugin_providers(self):
            yield "my-provider-plugin", (ecosystem_name, network_name, MyProvider)

    class MyEcosystem(Ethereum):
        pass

    ecosystem = MyEcosystem(name=ecosystem_name)
    network = MyNetwork(name=network_name, ecosystem=ecosystem)
    actual = list(network.providers)
    assert actual == [provider_name]  # And not "my-provider-plugin"


def test_providers_custom_network(project, custom_networks_config_dict, ethereum):
    with project.temp_config(**custom_networks_config_dict):
        network = ethereum.apenet
        actual = network.providers
        assert "node" in actual
        node = actual["node"]
        # NOTE: There was a bug where sometimes it would use the GethDev class
        #   and sometimes it would use the Node class. Node is what we want.
        assert "Node" in repr(node.func)
        assert "GethDev" not in repr(node.func)


def test_providers_custom_non_fork_network_does_not_use_fork_provider(
    mocker, project, custom_networks_config_dict, ethereum
):
    # NOTE: Have to a mock a Fork provider since none ship with Ape core.
    with project.temp_config(**custom_networks_config_dict):
        network = ethereum.apenet
        network.__dict__.pop("providers", None)  # de-cache

        # Setup mock fork provider.
        orig = network._get_plugin_providers
        network._get_plugin_providers = mocker.MagicMock()
        name = "foobar"

        class MyForkProvider:
            __module__ = "foobar.test"

        network._get_plugin_providers.return_value = iter(
            [(name, ("ethereum", "local", MyForkProvider))]
        )
        try:
            actual = network.providers
            assert name not in actual
        finally:
            network._get_plugin_providers = orig
            network.__dict__.pop("providers", None)  # de-cache


def test_is_local(ethereum):
    assert ethereum.local.is_local
    assert not ethereum.mainnet.is_local
    assert not ethereum.mainnet_fork.is_local


def test_is_fork(ethereum):
    assert not ethereum.local.is_fork
    assert not ethereum.mainnet.is_fork
    assert ethereum.mainnet_fork.is_fork


def test_is_dev(ethereum):
    assert ethereum.local.is_dev
    assert not ethereum.mainnet.is_dev
    assert ethereum.mainnet_fork.is_dev


def test_is_mainnet(ethereum):
    assert not ethereum.local.is_mainnet
    assert ethereum.mainnet.is_mainnet
    assert not ethereum.mainnet_fork.is_mainnet


def test_is_mainnet_from_config():
    """
    Simulate an EVM plugin with a weird named mainnet that properly
    configured it.
    """
    chain_id = 9191919191919919121177171
    ecosystem_name = "ismainnettest"
    network_name = "primarynetwork"
    network_type = create_network_type(chain_id, chain_id)

    class MyConfig(BaseEthereumConfig):
        primarynetwork: NetworkConfig = create_network_config(is_mainnet=True)

    class MyEcosystem(Ethereum):
        name: str = ecosystem_name

        @property
        def config(self):
            return MyConfig()

    ecosystem = MyEcosystem()
    network = network_type(name=network_name, ecosystem=ecosystem)
    assert network.is_mainnet


def test_explorer(networks):
    """
    Local network does not have an explorer, by default.
    """
    network = networks.ethereum.local
    network.__dict__.pop("explorer", None)  # Ensure not cached yet.
    assert network.explorer is None


def test_explorer_when_network_registered(networks, mocker):
    """
    Tests the simple flow of having the Explorer plugin register
    the networks it supports.
    """
    network = networks.ethereum.local
    network.__dict__.pop("explorer", None)  # Ensure not cached yet.
    name = "my-explorer"

    def explorer_cls(*args, **kwargs):
        res = mocker.MagicMock()
        res.name = name
        return res

    mock_plugin_explorers = mocker.patch(
        "ape.api.networks.NetworkAPI._plugin_explorers", new_callable=mock.PropertyMock
    )
    mock_plugin_explorers.return_value = [("my-example", ("ethereum", "local", explorer_cls))]
    assert network.explorer is not None
    assert network.explorer.name == name


def test_explorer_when_adhoc_network_supported(networks, mocker):
    """
    Tests the flow of when a chain is supported by an explorer
    but not registered in the plugin (API-flow).
    """
    network = networks.ethereum.local
    network.__dict__.pop("explorer", None)  # Ensure not cached yet.
    NAME = "my-explorer"

    class MyExplorer:
        name: str = NAME

        def __init__(self, *args, **kwargs):
            pass

        @classmethod
        def supports_chain(cls, chain_id):
            return True

    mock_plugin_explorers = mocker.patch(
        "ape.api.networks.NetworkAPI._plugin_explorers", new_callable=mock.PropertyMock
    )

    # NOTE: Ethereum is not registered at the plugin level, but is at the API level.
    mock_plugin_explorers.return_value = [
        ("my-example", ("some-other-ecosystem", "local", MyExplorer))
    ]
    assert network.explorer is not None
    assert network.explorer.name == NAME


def test_evm_chains_auto_forked_networks_exist(networks):
    # NOTE: Moonbeam networks exist in evmchains only; that is how Ape knows about them.
    assert isinstance(networks.moonbeam.moonriver_fork, ForkedNetworkAPI)



from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar

from ape.utils.basemodel import BaseInterfaceModel

ConvertedType = TypeVar("ConvertedType")


class ConverterAPI(BaseInterfaceModel, Generic[ConvertedType]):
    @abstractmethod
    def is_convertible(self, value: Any) -> bool:
        """
        Returns ``True`` if string value provided by ``value`` is convertible using
        :meth:`ape.api.convert.ConverterAPI.convert`.

        Args:
            value (Any): The value to check.

        Returns:
            bool: ``True`` when the given value can be converted.
        """

    @abstractmethod
    def convert(self, value: Any) -> ConvertedType:
        """
        Convert the given value to the type specified as the generic for this class.
        Implementations of this API must throw a :class:`~ape.exceptions.ConversionError`
        when the item fails to convert properly.

        Usage example::

            from ape import convert
            from ape.types import AddressType

            convert("1 gwei", int)
            # 1000000000

            convert("1 ETH", int)
            # 1000000000000000000

            convert("0x283Af0B28c62C092C9727F1Ee09c02CA627EB7F5", bytes)
            # HexBytes('0x283af0b28c62c092c9727f1ee09c02ca627eb7f5')

            convert("vitalik.eth", AddressType)  # with ape-ens plugin installed
            # '0xd8dA6BF26964aF9D7eEd9e03E53415D37aA96045'

        """

    @property
    def name(self) -> str:
        """
        The calculated name of the converter class.
        Typically, it is the lowered prefix of the class without
        the "Converter" or "Conversions" suffix.
        """
        class_name = self.__class__.__name__
        name = class_name.replace("Converter", "").replace("Conversions", "")
        return name.lower()


class ConvertibleAPI(ABC):
    """
    Use this base-class mixin if you want your custom class to be convertible to a more basic type
    without having to register a converter plugin for it.
    """

    @abstractmethod
    def is_convertible(self, to_type: type) -> bool:
        """
        Returns ``True`` if ``self`` can be converted to ``to_type``.
        """

    @abstractmethod
    def convert_to(self, to_type: type) -> Any:
        """
        Convert ``self`` to the given type. Implementing classes _should_ raise ``ConversionError`` if not convertible.
        Ape's conversion system will **only** attempt to convert classes where ``.is_convertible()`` returns ``True``.
        """



import pytest
from ethpm_types import ContractType

from ape import Contract
from ape.contracts import ContractContainer, ContractInstance
from ape.exceptions import (
    ArgumentsLengthError,
    MethodNonPayableError,
    MissingDeploymentBytecodeError,
    NetworkError,
    ProjectError,
)
from ape_ethereum.ecosystem import ProxyType
from tests.conftest import explorer_test


def test_deploy(
    not_owner,
    contract_container,
    networks_connected_to_tester,
    clean_contract_caches,
):
    contract = contract_container.deploy(4, sender=not_owner, something_else="IGNORED")
    assert contract.txn_hash
    assert contract.myNumber() == 4

    # Verify can reload same contract from cache
    contract_from_cache = Contract(contract.address)
    assert contract_from_cache.contract_type == contract.contract_type
    assert contract_from_cache.address == contract.address
    assert contract_from_cache.txn_hash == contract.txn_hash


def test_deploy_wrong_number_of_arguments(
    not_owner,
    contract_container,
    networks_connected_to_tester,
    clean_contract_caches,
):
    expected = (
        r"The number of the given arguments \(0\) does not match what is defined in the "
        r"ABI:\n\n\t.*__init__\(uint256 num\).*"
    )
    with pytest.raises(ArgumentsLengthError, match=expected):
        contract_container.deploy(sender=not_owner)


def test_deploy_and_publish_local_network(owner, contract_container):
    with pytest.raises(ProjectError, match="Can only publish deployments on a live network"):
        contract_container.deploy(0, sender=owner, publish=True)


def test_deploy_and_publish_live_network_no_explorer(owner, contract_container, dummy_live_network):
    dummy_live_network.__dict__["explorer"] = None
    expected_message = "Unable to publish contract - no explorer plugin installed."
    with pytest.raises(NetworkError, match=expected_message):
        contract_container.deploy(0, sender=owner, publish=True, required_confirmations=0)


@explorer_test
def test_deploy_and_publish(
    owner, contract_container, dummy_live_network_with_explorer, mock_explorer
):
    contract = contract_container.deploy(0, sender=owner, publish=True, required_confirmations=0)
    mock_explorer.publish_contract.assert_called_once_with(contract.address)


@explorer_test
def test_deploy_and_not_publish(
    owner, contract_container, dummy_live_network_with_explorer, mock_explorer
):
    contract_container.deploy(0, sender=owner, publish=False, required_confirmations=0)
    assert not mock_explorer.call_count


def test_deploy_privately(owner, contract_container):
    deploy_0 = owner.deploy(contract_container, 3, private=True)
    assert isinstance(deploy_0, ContractInstance)

    deploy_1 = contract_container.deploy(3, sender=owner, private=True)
    assert isinstance(deploy_1, ContractInstance)


@pytest.mark.parametrize("bytecode", (None, {}, {"bytecode": "0x"}))
def test_deploy_no_deployment_bytecode(owner, bytecode):
    """
    https://github.com/ApeWorX/ape/issues/1904
    """
    expected = (
        r"Cannot deploy: contract 'Apes' has no deployment-bytecode\. "
        r"Are you attempting to deploy an interface\?"
    )
    contract_type = ContractType.model_validate(
        {"abi": [], "contractName": "Apes", "deploymentBytecode": bytecode}
    )
    contract = ContractContainer(contract_type)
    with pytest.raises(MissingDeploymentBytecodeError, match=expected):
        contract.deploy(sender=owner)


def test_deploy_sending_funds_to_non_payable_constructor(project, owner):
    with pytest.raises(
        MethodNonPayableError,
        match=r"Sending funds to a non-payable constructor\.",
    ):
        project.SolidityContract.deploy(1, sender=owner, value="1 ether")


def test_deployments(owner, eth_tester_provider, project):
    initial_deployed_contract = project.VyperContract.deploy(10000000, sender=owner)
    actual = project.VyperContract.deployments[-1].address
    expected = initial_deployed_contract.address
    assert actual == expected


def test_deploy_proxy(owner, vyper_contract_instance, project, chain, eth_tester_provider):
    target = vyper_contract_instance.address
    proxy = project.SimpleProxy.deploy(target, sender=owner)

    # Ensure we can call both proxy and target methods on it.
    assert proxy.implementation  # No attr err
    assert proxy.myNumber  # No attr err

    # Ensure caching works.
    assert proxy.address in chain.contracts.contract_types
    assert proxy.address in chain.contracts.proxy_infos

    # Show the cached proxy info is correct.
    proxy_info = chain.contracts.proxy_infos[proxy.address]
    assert proxy_info.target == target
    assert proxy_info.type == ProxyType.Delegate

    # Show we get the implementation contract type using the proxy address
    re_contract = chain.contracts.instance_at(proxy.address)
    assert re_contract.contract_type == proxy.contract_type

    # Show proxy methods are not available on target alone.
    target = chain.contracts.instance_at(proxy_info.target)
    assert target.myNumber  # No attr err
    with pytest.raises(AttributeError):
        _ = target.implementation


def test_source_path_in_project(project_with_contract):
    contract = project_with_contract.contracts["Contract"]
    contract_container = project_with_contract.get_contract("Contract")
    expected = project_with_contract.path / contract.source_id
    assert contract_container.source_path.is_file()
    assert contract_container.source_path == expected


def test_source_path_out_of_project(solidity_contract_instance, project):
    solidity_contract_instance.base_path = None
    with project.isolate_in_tempdir(chdir=True):
        assert not solidity_contract_instance.source_path


def test_encode_constructor_input(contract_container, calldata):
    constructor = contract_container.constructor
    actual = constructor.encode_input(222)
    expected = calldata[4:]  # Strip off setNumber() method ID
    assert actual == expected


def test_decode_constructor_input(contract_container, calldata):
    constructor = contract_container.constructor
    constructor_calldata = calldata[4:]  # Strip off setNumber() method ID
    actual = constructor.decode_input(constructor_calldata)
    expected = "constructor(uint256)", {"num": 222}
    assert actual == expected


def test_decode_input(contract_container, calldata):
    actual = contract_container.decode_input(calldata)
    expected = "setNumber(uint256)", {"num": 222}
    assert actual == expected


def test_declare(contract_container, sender):
    receipt = contract_container.declare(sender=sender)
    assert not receipt.failed


def test_source_id(contract_container):
    actual = contract_container.source_id
    expected = contract_container.contract_type.source_id
    # Is just a pass-through (via extras-model), but making sure it works.
    assert actual == expected


def test_at(vyper_contract_instance, project):
    instance = project.VyperContract.at(vyper_contract_instance.address)
    assert instance == vyper_contract_instance


def test_at_fetch_from_explorer_false(
    project_with_contract, mock_explorer, eth_tester_provider, owner
):
    # Grab the container - note: this always compiles!
    container = project_with_contract.Contract
    instance = container.deploy(sender=owner)

    # Hack off the fact that it was compiled.
    project_with_contract.clean()

    # Simulate having an explorer plugin installed (e.g. ape-etherscan).
    eth_tester_provider.network.__dict__["explorer"] = mock_explorer

    # Attempt to create an instance. It should NOT use the explorer at all!
    instance2 = container.at(instance.address, fetch_from_explorer=False)

    assert instance == instance2
    # Ensure explorer was not used at all.
    assert mock_explorer.get_contract_type.call_count == 0

    # Clean up test.
    eth_tester_provider.network.__dict__.pop("explorer")



from importlib import import_module

from ape.plugins import Config, QueryPlugin, register


@register(Config)
def config_class():
    from ape_cache.config import CacheConfig

    return CacheConfig


@register(QueryPlugin)
def query_engines():
    query = import_module("ape_cache.query")
    return query.CacheQueryProvider


def __getattr__(name):
    if name == "CacheQueryProvider":
        module = import_module("ape_cache.query")
        return module.CacheQueryProvider

    elif name == "CacheConfig":
        module = import_module("ape_cache.config")
        return module.CacheConfig

    else:
        raise AttributeError(name)


__all__ = [
    "CacheConfig",
    "CacheQueryProvider",
]



import signal
import threading

if threading.current_thread() is threading.main_thread():
    # If we are in the main thread, we can safely set the signal handler
    signal.signal(signal.SIGINT, lambda s, f: _sys.exit(130))

import sys as _sys

__all__ = [
    "Contract",
    "Project",  # So you can load other projects
    "accounts",
    "chain",
    "compilers",
    "config",
    "convert",
    "fixture",
    "networks",
    "project",
    "reverts",
]


def __getattr__(name: str):
    if name not in __all__:
        raise AttributeError(name)

    elif name == "reverts":
        from ape.pytest.contextmanagers import RevertsContextManager

        return RevertsContextManager

    elif name == "fixture":
        from ape.pytest.fixtures import fixture

        return fixture

    else:
        from ape.utils.basemodel import ManagerAccessMixin as access

        if name == "Contract":
            return access.chain_manager.contracts.instance_at

        elif name == "Project":
            return access.Project

        elif name == "convert":
            return access.conversion_manager.convert

        # The rest are managers; we can derive the name.
        key = name
        if name == "project":
            key = "local_project"
        elif name.endswith("s"):
            key = f"{name[:-1]}_manager"
        else:
            key = f"{key}_manager"

        return getattr(access, key)



import io
import os
import re
from pathlib import Path

import pytest

from ape.logging import logger
from ape.pytest.fixtures import PytestApeFixtures
from tests.conftest import GETH_URI, geth_process_test
from tests.integration.cli.utils import skip_projects_except

BASE_PROJECTS_PATH = Path(__file__).parent / "projects"
TOKEN_B_GAS_REPORT = r"""
 +TokenB Gas

  Method +Times called +Min. +Max. +Mean +Median
 ─+
  __init__ +\d +\d+ + \d+ + \d+ + \d+
  balanceOf +\d +\d+ + \d+ + \d+ + \d+
  transfer +\d +\d+ + \d+ + \d+ + \d+
"""
EXPECTED_GAS_REPORT = rf"""
 +VyperContract Gas

  Method +Times called +Min. +Max. +Mean +Median
 ─+
  __init__ +\d +\d+ + \d+ + \d+ + \d+
  fooAndBar +\d +\d+ + \d+ + \d+ + \d+
  myNumber +\d +\d+ + \d+ + \d+ + \d+
  setAddress +\d +\d+ + \d+ + \d+ + \d+
  setNumber +\d +\d+ + \d+ + \d+ + \d+

 +TokenA Gas

  Method +Times called +Min. +Max. +Mean +Median
 ─+
  __init__ +\d +\d+ + \d+ + \d+ + \d+
  balanceOf +\d +\d+ + \d+ + \d+ + \d+
  transfer +\d +\d+ + \d+ + \d+ + \d+
{TOKEN_B_GAS_REPORT}
"""
GETH_LOCAL_CONFIG = f"""
node:
  ethereum:
    local:
      uri: {GETH_URI}
"""


def filter_expected_methods(*methods_to_remove: str) -> str:
    expected = EXPECTED_GAS_REPORT
    for name in methods_to_remove:
        line = f"\n  {name} +\\d +\\d+ + \\d+ + \\d+ + \\d+"
        expected = expected.replace(line, "")

    return expected


@pytest.fixture(autouse=True, scope="session")
def path_check():
    # Fix annoying issue where path cwd doesn't exist
    # 'Something something' temp path.
    try:
        os.getcwd()
    except Exception:
        os.chdir(f"{Path(__file__).parent}")


@pytest.fixture(autouse=True)
def load_dependencies(integ_project):
    """Ensure these are loaded before setting up pytester."""
    integ_project.dependencies.install()


@pytest.fixture
def setup_pytester(pytester, owner):
    # Mine to a new block so we are not capturing old transactions
    # in the tests.
    owner.transfer(owner, 0)

    def setup(project):
        project_path = BASE_PROJECTS_PATH / project.path.name
        tests_path = project_path / "tests"

        # Assume all tests should pass
        num_passes = 0
        num_failed = 0
        test_files = {}
        if tests_path.is_dir():
            for file_path in tests_path.iterdir():
                if not file_path.name.startswith("test_") or file_path.suffix != ".py":
                    continue

                content = file_path.read_text(encoding="utf8")
                test_files[file_path.name] = content
                num_passes += len(
                    [
                        x
                        for x in content.splitlines()
                        if x.lstrip().startswith("def test_") and not x.startswith("def test_fail_")
                    ]
                )
                num_failed += len(
                    [x for x in content.splitlines() if x.startswith("def test_fail_")]
                )

            pytester.makepyfile(**test_files)

        # Make other files
        def _make_all_files(base: Path, prefix: Path | None = None):
            if not base.is_dir():
                return

            for file in base.iterdir():
                if file.is_dir() and file.name != "tests":
                    _make_all_files(file, prefix=Path(file.name))
                elif file.is_file() and file.suffix not in (".sol", ".vy"):
                    name = (prefix / file.name).as_posix() if prefix else file.name

                    if name == "ape-config.yaml":
                        # Hack in in-memory overrides for testing purposes.
                        text = str(project.config)
                    else:
                        text = file.read_text(encoding="utf8")

                    src = {name: text.splitlines()}
                    pytester.makefile(file.suffix, **src)

        _make_all_files(project_path)

        # Check for a conftest.py
        conftest = tests_path / "conftest.py"
        if conftest.is_file():
            pytester.makeconftest(conftest.read_text())

        # Returns expected number of passing tests.
        return num_passes, num_failed

    return setup


def run_gas_test(
    result, expected_passed: int, expected_failed: int, expected_report: str = EXPECTED_GAS_REPORT
):
    try:
        result.assert_outcomes(passed=expected_passed, failed=expected_failed)
    except AssertionError:
        pytest.fail("\n".join(result.outlines))

    gas_header_line_index = None
    for index, line in enumerate(result.outlines):
        if "Gas Profile" in line:
            gas_header_line_index = index

    assert gas_header_line_index is not None, "'Gas Profile' not in output."
    expected = [x for x in expected_report.rstrip().split("\n")[1:]]
    start_index = gas_header_line_index + 1
    end_index = start_index + len(expected)
    actual = [x.rstrip() for x in result.outlines[start_index:end_index]]
    assert "WARNING: No gas usage data found." not in actual, "Gas data missing!"

    actual_len = len(actual)
    expected_len = len(expected)

    if actual_len > expected_len:
        remainder = "\n".join(actual[expected_len:])
        pytest.fail(f"Actual contains more than expected:\n{remainder}")
    elif expected_len > actual_len:
        remainder = "\n".join(expected[actual_len:])
        pytest.fail(f"Expected contains more than actual:\n{remainder}")

    for actual_line, expected_line in zip(actual, expected, strict=True):
        message = f"'{actual_line}' does not match pattern '{expected_line}'."
        assert re.match(expected_line, actual_line), message


@skip_projects_except("test", "with-contracts")
def test_test(setup_pytester, integ_project, pytester, eth_tester_provider):
    _ = eth_tester_provider  # Ensure using EthTester for this test.
    passed, failed = setup_pytester(integ_project)

    with logger.at_level("DEBUG"):
        result = pytester.runpytest_subprocess("-n", "0", timeout=120)

    outcomes = result.parseoutcomes()
    assert "failed" not in outcomes if failed == 0 else outcomes["failed"] == failed
    if integ_project.name != "test":
        assert outcomes["passed"] == passed
    # else: too many parametrized tests to calculate. "No fails" is good enough.


@skip_projects_except("with-contracts")
def test_uncaught_txn_err(setup_pytester, integ_project, pytester, eth_tester_provider):
    _ = eth_tester_provider  # Ensure using EthTester for this test.
    setup_pytester(integ_project)
    result = pytester.runpytest_subprocess("-n", "0", timeout=120)
    expected = """
    contract_in_test.setNumber(5, sender=owner)
E   ape.exceptions.ContractLogicError: Transaction failed.
    """.strip()
    assert expected in str(result.stdout)


@skip_projects_except("with-contracts")
def test_show_internal(setup_pytester, integ_project, pytester, eth_tester_provider):
    _ = eth_tester_provider  # Ensure using EthTester for this test.
    setup_pytester(integ_project)
    result = pytester.runpytest_subprocess("-n", "0", "--show-internal")
    expected = """
    raise vm_err from err
E   ape.exceptions.ContractLogicError: Transaction failed.
    """.strip()
    assert expected in str(result.stdout)


@skip_projects_except("test", "with-contracts")
def test_isolation_disabled(setup_pytester, integ_project, pytester, eth_tester_provider):
    # check the disable isolation option actually disables built-in isolation
    _ = eth_tester_provider  # Ensure using EthTester for this test.
    setup_pytester(integ_project)
    result = pytester.runpytest_subprocess("--disable-isolation", "--setup-show", "-n", "0")
    assert "F _function_isolation" not in "\n".join(result.outlines)


@skip_projects_except("test")
def test_verbosity(runner, ape_cli):
    """
    Tests again an issue where `ape test -v debug` would fail because of
    an invalid type check from click; only appeared in `ape test` command
    for some reason.
    """
    # NOTE: Only using `--fixtures` flag to avoid running tests (just prints fixtures).
    cmd = ("test", "--verbosity", "DEBUG", "--fixtures")
    result = runner.invoke(ape_cli, cmd, catch_exceptions=False)
    assert result.exit_code == 0, result.output


@skip_projects_except("test")
@pytest.mark.parametrize("v_arg", ("-v", "-vv", "-vvv"))
def test_vvv(runner, ape_cli, integ_project, v_arg):
    """
    Showing you can somehow use pytest's -v flag without
    messing up Ape.
    """
    here = integ_project.path
    os.chdir(integ_project.path)
    name = f"test_{v_arg.replace('-', '_')}"

    TEST = f"""
    def {name}():
        assert True
    """.lstrip()

    # Have to create a new test each time to avoid .pycs issues
    new_test_file = integ_project.tests_folder / f"{name}.py"
    new_test_file.write_text(TEST)

    try:
        # NOTE: v_arg purposely at the end because testing doesn't interfere
        #   with click's option parsing "requires value" error.
        result = runner.invoke(
            ape_cli,
            ("test", f"tests/{new_test_file.name}::{name}", v_arg),
            catch_exceptions=False,
        )
    finally:
        new_test_file.unlink(missing_ok=True)
        os.chdir(here)

    assert result.exit_code == 0, result.output
    # Prove `-vvv` worked via the output.
    # It shows PASSED instead of the little green dot.
    assert "PASSED" in result.output


@skip_projects_except("test", "with-contracts")
def test_fixture_docs(setup_pytester, integ_project, pytester, eth_tester_provider):
    _ = eth_tester_provider  # Ensure using EthTester for this test.
    result = pytester.runpytest_subprocess("-q", "--fixtures", "-n", "0")
    actual = "\n".join(result.outlines)

    # 'accounts', 'networks', 'chain', and 'project' (etc.)
    fixtures = [prop for n, prop in vars(PytestApeFixtures).items() if not n.startswith("_")]
    for fixture in fixtures:
        # The doc str of the fixture shows in the CLI output
        for doc_str in fixture.__doc__.splitlines():
            assert doc_str.strip() in actual


@skip_projects_except("with-contracts")
def test_gas_flag_when_not_supported(setup_pytester, integ_project, pytester, eth_tester_provider):
    _ = eth_tester_provider  # Ensure using EthTester for this test.
    setup_pytester(integ_project)
    path = f"{integ_project.path}/tests/test_contract.py"
    path_w_test = f"{path}::test_contract_interaction_in_tests"
    result = pytester.runpytest(path_w_test, "--gas", "-n", "0")
    actual = "\n".join(result.outlines)
    expected = (
        "Provider 'test' does not support transaction tracing. "
        "The gas profile is limited to receipt-level data."
    )
    assert expected in actual


@pytest.mark.flaky(reruns=2)
@geth_process_test
@skip_projects_except("geth")
def test_gas_flag_in_tests(geth_provider, setup_pytester, integ_project, pytester, owner):
    owner.transfer(owner, "1 wei")  # Do this to force a clean slate.
    passed, failed = setup_pytester(integ_project)
    result = pytester.runpytest_subprocess("--gas", "--network", "ethereum:local:node", "-n", "0")
    run_gas_test(result, passed, failed)


@pytest.mark.flaky(reruns=2)
@geth_process_test
@skip_projects_except("geth")
def test_gas_flag_set_in_config(
    geth_provider, setup_pytester, integ_project, pytester, geth_account
):
    geth_account.transfer(geth_account, "1 wei")  # Force a clean block.
    cfg = integ_project.config.model_dump(by_alias=True, mode="json")
    cfg["test"]["gas"] = {"reports": ["terminal"]}
    with integ_project.temp_config(**cfg):
        passed, failed = setup_pytester(integ_project)
        result = pytester.runpytest_subprocess("--network", "ethereum:local:node", "-n", "0")

    run_gas_test(result, passed, failed)


@pytest.mark.flaky(reruns=2)
@geth_process_test
@skip_projects_except("geth")
def test_gas_when_estimating(geth_provider, setup_pytester, integ_project, pytester, geth_account):
    """
    Shows that gas reports still work when estimating gas.
    """
    cfg = integ_project.config.model_dump(by_alias=True, mode="json")
    cfg["test"]["gas"] = {"reports": ["terminal"]}
    geth_account.transfer(geth_account, "1 wei")  # Force a clean block.
    with integ_project.temp_config(**cfg):
        passed, failed = setup_pytester(integ_project)
        result = pytester.runpytest_subprocess("-n", "0", timeout=120)
        run_gas_test(result, passed, failed)


@pytest.mark.flaky(reruns=2)
@geth_process_test
@skip_projects_except("geth")
def test_gas_flag_exclude_using_cli_option(
    geth_provider, setup_pytester, integ_project, pytester, geth_account
):
    geth_account.transfer(geth_account, "1 wei")  # Force a clean block.
    passed, failed = setup_pytester(integ_project)
    # NOTE: Includes both a mutable and a view method.
    expected = filter_expected_methods("fooAndBar", "myNumber")
    # Also ensure can filter out whole class
    expected = expected.replace(TOKEN_B_GAS_REPORT, "")
    result = pytester.runpytest_subprocess(
        "--gas",
        "--gas-exclude",
        "*:fooAndBar,*:myNumber,tokenB:*",
        "--network",
        "ethereum:local:node",
        "-n",
        "0",
    )
    run_gas_test(result, passed, failed, expected_report=expected)


@pytest.mark.flaky(reruns=2)
@geth_process_test
@skip_projects_except("geth")
def test_gas_flag_exclusions_set_in_config(
    geth_provider, setup_pytester, integ_project, pytester, geth_account
):
    # Attempt 3 times to bump the block before giving up.
    for _ in range(3):
        try:
            geth_account.transfer(geth_account, "1 wei")  # Force a clean block.
        except Exception as err:
            # Sometimes tests are weird, since running async.
            # So give this a chance a few times before bailing out...
            logger.error(f"Failure bumping block: {err}")
            continue

    # NOTE: Includes both a mutable and a view method.
    expected = filter_expected_methods("fooAndBar", "myNumber")
    # Also ensure can filter out whole class
    expected = expected.replace(TOKEN_B_GAS_REPORT, "")
    cfg = integ_project.config.model_dump(by_alias=True, mode="json")
    cfg["test"]["gas"] = {
        "exclude": [
            {"method_name": "fooAndBar"},
            {"method_name": "myNumber"},
            {"contract_name": "TokenB"},
        ]
    }
    with integ_project.temp_config(**cfg):
        passed, failed = setup_pytester(integ_project)
        result = pytester.runpytest_subprocess(
            "--gas", "--network", "ethereum:local:node", "-n", "0"
        )
        run_gas_test(result, passed, failed, expected_report=expected)


@pytest.mark.flaky(reruns=2)
@geth_process_test
@skip_projects_except("geth")
def test_gas_flag_excluding_contracts(
    geth_provider, setup_pytester, integ_project, pytester, geth_account
):
    geth_account.transfer(geth_account, "1 wei")  # Force a clean block.
    passed, failed = setup_pytester(integ_project)
    result = pytester.runpytest_subprocess(
        "--gas",
        "--gas-exclude",
        "VyperContract,TokenA",
        "--network",
        "ethereum:local:node",
        "-n",
        "0",
    )
    run_gas_test(result, passed, failed, expected_report=TOKEN_B_GAS_REPORT)


@pytest.mark.flaky(reruns=2)
@geth_process_test
@skip_projects_except("geth")
def test_coverage(geth_provider, setup_pytester, integ_project, pytester, geth_account):
    """
    Ensures the --coverage flag works.
    For better coverage tests, see ape-vyper because the Vyper
    plugin is what implements the `trace_source()` method which does the bulk
    of the coverage work.
    """
    geth_account.transfer(geth_account, "1 wei")  # Force a clean block.
    passed, failed = setup_pytester(integ_project)
    result = pytester.runpytest_subprocess(
        "--coverage", "--show-internal", "--network", "ethereum:local:node", "-n", "0"
    )
    result.assert_outcomes(passed=passed, failed=failed)


@skip_projects_except("with-contracts")
def test_interactive(eth_tester_provider, integ_project, pytester, monkeypatch):
    secret = "__ 123 super secret 123 __"
    test = f"""
def test_fails():
    foo = "{secret}"
    raise Exception("__FAIL__")
"""
    pytester.makepyfile(test)
    stdin = "print(foo)\nexit\n"
    monkeypatch.setattr("sys.stdin", io.StringIO(stdin))
    result = pytester.runpytest_subprocess("--interactive", "-s", "-n", "0")
    result.assert_outcomes(failed=1)
    actual = str(result.stdout)
    assert secret in actual
    assert "__FAIL__" in actual


@skip_projects_except("with-contracts")
def test_watch(mocker, integ_project, runner, ape_cli):
    runner_patch = mocker.patch("ape_test._cli._run_with_observer")

    # Only passing `-s` so we have an extra arg to test.
    result = runner.invoke(ape_cli, ("test", "--watch", "-s"))
    assert result.exit_code == 0

    runner_patch.assert_called_once_with((Path("contracts"), Path("tests")), 0.5, "-s")


@skip_projects_except("with-contracts")
def test_project_option(integ_project, ape_cli, runner):
    _ = integ_project  # NOTE: Not actually used, but avoid running across all projects.

    # NOTE: Using isolated filesystem so that
    with runner.isolated_filesystem():
        # Set up a project
        project_name = "test-token-project"
        project_dir = Path.cwd() / project_name
        tests_dir = project_dir / "tests"
        tests_dir.mkdir(parents=True)

        # Set up a test
        test_file = tests_dir / "test_project_option.py"
        test_text = """
def test_project_option():
    assert True
""".lstrip()
        test_file.write_text(test_text)

        result = runner.invoke(ape_cli, ("test", "--project", f"./{project_name}"))
        assert "1 passed" in result.output
        assert result.exit_code == 0



import pytest

from ape.exceptions import ContractNotFoundError
from tests.conftest import geth_process_test


@geth_process_test
def test_get_proxy_from_explorer(
    mock_explorer,
    create_mock_sepolia,
    project,
    geth_account,
    geth_provider,
    chain,
):
    """
    Simulated when you get a contract from Etherscan for the first time
    but that contract is a proxy. We expect both proxy and target ABIs
    to be cached under the proxy's address.
    """
    target_contract = geth_account.deploy(project.VyperContract, 10011339315)
    proxy_contract = geth_account.deploy(project.SafeProxy, target_contract.address)

    # Ensure both of these are not cached so we have to rely on our fake explorer.
    del chain.contracts[target_contract.address]
    del chain.contracts[proxy_contract.address]
    # Sanity check.
    with pytest.raises(ContractNotFoundError):
        _ = chain.contracts.instance_at(proxy_contract.address)

    def get_contract_type(address, *args, **kwargs):
        # Mock etherscan backend.
        if address == target_contract.address:
            return target_contract.contract_type
        elif address == proxy_contract.address:
            return proxy_contract.contract_type

        raise ValueError("Fake explorer only knows about proxy and target contracts.")

    with create_mock_sepolia() as network:
        # Set up our network to use our fake explorer.
        mock_explorer.get_contract_type.side_effect = get_contract_type
        network.__dict__["explorer"] = mock_explorer

        # Typical flow: user attempts to get an un-cached contract type from Etherscan.
        # That contract may be a proxy, in which case we should get a type
        # w/ both proxy ABIs and the target ABIs.
        contract_from_explorer = chain.contracts.instance_at(proxy_contract.address)

        network.__dict__.pop("explorer", None)

        # Ensure we can call proxy methods!
        assert contract_from_explorer.masterCopy  # No attr error!

        # Ensure we can call target methods!
        assert contract_from_explorer.myNumber  # No attr error!



def __getattr__(name: str):
    if name == "AccountManager":
        from ape.managers.accounts import AccountManager

        return AccountManager

    elif name == "ChainManager":
        from ape.managers.chain import ChainManager

        return ChainManager

    elif name == "CompilerManager":
        from ape.managers.compilers import CompilerManager

        return CompilerManager

    elif name == "ConfigManager":
        from ape.managers.config import ConfigManager

        return ConfigManager

    elif name == "ConversionManager":
        from ape.managers.converters import ConversionManager

        return ConversionManager

    elif name == "NetworkManager":
        from ape.managers.networks import NetworkManager

        return NetworkManager

    elif name == "PluginManager":
        from ape.managers.plugins import PluginManager

        return PluginManager

    elif name == "ProjectManager":
        from ape.managers.project import ProjectManager

        return ProjectManager

    elif name == "QueryManager":
        from ape.managers.query import QueryManager

        return QueryManager

    else:
        raise AttributeError(name)



import threading
import time
from collections.abc import Iterable
from datetime import datetime, timedelta
from functools import cached_property
from pathlib import Path
from subprocess import run as run_subprocess

from watchdog import events
from watchdog.observers import Observer

from ape.logging import logger

# Copied from https://github.com/olzhasar/pytest-watcher/blob/master/pytest_watcher/watcher.py
trigger_lock = threading.Lock()
trigger = None


def run_with_observer(watch_folders: Iterable[Path], watch_delay: float, *pytest_arg_ls: str):
    event_handler = _create_event_handler()
    observer = _create_observer()

    for folder in watch_folders:
        if folder.is_dir():
            observer.schedule(event_handler, folder, recursive=True)
        else:
            logger.warning(f"Folder '{folder}' doesn't exist or isn't a folder.")

    observer.start()

    try:
        _run_ape_test(*pytest_arg_ls)
        while True:
            _run_main_loop(watch_delay, *pytest_arg_ls)

    finally:
        observer.stop()
        observer.join()


def emit_trigger():
    """
    Emits trigger to run pytest
    """

    global trigger

    with trigger_lock:
        trigger = datetime.now()


class EventHandler(events.FileSystemEventHandler):
    EVENTS_WATCHED = (
        events.EVENT_TYPE_CREATED,
        events.EVENT_TYPE_DELETED,
        events.EVENT_TYPE_MODIFIED,
        events.EVENT_TYPE_MOVED,
    )

    def dispatch(self, event: events.FileSystemEvent) -> None:
        if event.event_type in self.EVENTS_WATCHED:
            self.process_event(event)

    @cached_property
    def _extensions_to_watch(self) -> list[str]:
        from ape.utils.basemodel import ManagerAccessMixin as access

        return [".py", *access.compiler_manager.registered_compilers.keys()]

    def _is_path_watched(self, filepath: str) -> bool:
        """
        Check if file should trigger pytest run
        """
        return any(map(filepath.endswith, self._extensions_to_watch))

    def process_event(self, event: events.FileSystemEvent) -> None:
        if self._is_path_watched(event.src_path):
            emit_trigger()


def _run_ape_test(*pytest_args):
    return run_subprocess(["ape", "test", *[f"{a}" for a in pytest_args]])


def _run_main_loop(delay: float, *pytest_args: str) -> None:
    global trigger

    now = datetime.now()
    if trigger and now - trigger > timedelta(seconds=delay):
        _run_ape_test(*pytest_args)

        with trigger_lock:
            trigger = None

    time.sleep(delay)


def _create_event_handler():
    # Abstracted for testing purposes.
    return EventHandler()


def _create_observer():
    # Abstracted for testing purposes.
    return Observer()



import pytest

from ape.managers.converters import HexAddressConverter, IntAddressConverter
from ape.types.address import AddressType


def test_convert_keyfile_account_to_address(convert, keyfile_account):
    actual = convert(keyfile_account, AddressType)
    assert actual == keyfile_account.address


def test_convert_test_account_to_address(convert, owner):
    actual = convert(owner, AddressType)
    assert actual == owner.address


def test_convert_address_to_int(convert, owner):
    actual = convert(owner.address, int)
    assert actual == int(owner.address, 16)


def test_convert_account_to_int(convert, owner):
    actual = convert(owner, int)
    assert actual == int(owner.address, 16)


def test_convert_0_to_address(convert, zero_address):
    assert convert(0, AddressType) == zero_address


class TestHexAddressConverter:
    @pytest.fixture(scope="class")
    def converter(self):
        return HexAddressConverter()

    def test_is_convertible_hex_str(self, converter):
        assert not converter.is_convertible("0x123")

    def test_is_convertible_address(self, converter, owner):
        # Is already an address!
        assert not converter.is_convertible(str(owner.address))

    def test_convert_not_canonical_address(self, converter):
        actual = converter.convert("0x0ffffffaaaaaaaabbbbbbb333337777eeeeeee00")
        expected = "0x0fFFfffAaAaAaAaBBBbBbb333337777eeeeeEe00"
        assert actual == expected


class TestIntAddressConverter:
    @pytest.fixture(scope="class")
    def converter(self):
        return IntAddressConverter()

    def test_is_convertible(self, converter, owner):
        int_address = int(owner.address, 16)
        assert converter.is_convertible(int_address)

    def test_is_convertible_random_int(self, converter):
        assert converter.is_convertible(0)

    @pytest.mark.parametrize("val", (0, 1))
    def test_convert_simple_int(self, converter, val, zero_address):
        actual = converter.convert(val)
        expected = f"{zero_address[:-1]}{val}"
        assert actual == expected
